from copy import deepcopy
from typing import Any, Callable, Dict, List, Optional

from supervisely.app import DataJson, StateJson
from supervisely.app.widgets import Widget


class ClassBalance(Widget):
    """Visualizes class distribution / balance of input data."""

    class Routes:
        """Callback route names used by the widget frontend to notify Python."""
        CLICK = "class_balance_clicked_cb"

    def __init__(
        self,
        segments: Optional[List[Dict]] = [],
        rows_data: Optional[List[Dict]] = [],
        slider_data: Optional[Dict[str, List]] = {},
        max_value: Optional[int] = None,
        max_height: Optional[int] = 350,
        rows_height: Optional[int] = 100,
        selectable: Optional[bool] = True,
        collapsable: Optional[bool] = False,
        clickable_name: Optional[bool] = False,
        clickable_segment: Optional[bool] = False,
        widget_id: Optional[str] = None,
    ):
        """
        :param segments: List of segments (name, key, color).
        :type segments: List[Dict]
        :param rows_data: List of row dicts (name, total, segments).
        :type rows_data: List[Dict]
        :param slider_data: Per-class slider/preview data.
        :type slider_data: Dict[str, List]
        :param max_value: Max value for scale.
        :type max_value: Optional[int]
        :param max_height: Max height in pixels.
        :type max_height: Optional[int]
        :param rows_height: Row height in pixels.
        :type rows_height: Optional[int]
        :param selectable: Enable selection.
        :type selectable: Optional[bool]
        :param collapsable: Enable collapse.
        :type collapsable: Optional[bool]
        :param clickable_name: Enable name click.
        :type clickable_name: Optional[bool]
        :param clickable_segment: Enable segment click.
        :type clickable_segment: Optional[bool]
        :param widget_id: Unique widget identifier.
        :type widget_id: Optional[str]

        :Usage Example:

            .. code-block:: python

                from supervisely.app.widgets import ClassBalance
                segments = [{"name": "train", "key": "train", "color": "#1892f8"}]
                rows_data = [{"name": "class1", "total": 100, "segments": {"train": 100}}]
                cb = ClassBalance(segments=segments, rows_data=rows_data, collapsable=True)
        """
        self._segments = segments
        self._rows_data = rows_data
        self._slider_data = slider_data
        self._max_value = max_value
        self._max_height = f"{max_height}px"
        self._rows_height = f"{rows_height}px"
        self._selectable = selectable
        self._collapsable = collapsable
        self._clickable_name = clickable_name
        self._clickable_segment = clickable_segment
        self._click_handled = False

        if self._max_value is None:
            if len(self._rows_data) != 0:
                self._max_value = max([curr_row["total"] for curr_row in self._rows_data])
            else:
                self._max_value = 0

        for curr_row in self._rows_data:
            if curr_row.get("segments") is None:
                curr_row["segments"] = {}

        super().__init__(widget_id=widget_id, file_path=__file__)

    def get_json_data(self) -> Dict[str, Any]:
        """Returns dictionary with widget data, which defines the appearance and behavior of the widget.

        Dictionary contains the following fields:
            - content: Dictionary with the following fields:
                - maxValue: Maximum value of the widget.
                - segments: List of segments to be displayed in the widget.
                - rows: List of rows to be displayed in the widget.
            - options: Dictionary with the following fields:
                - selectable: If True, the widget will be selectable.
                - collapsable: If True, the widget will be collapsable.
                - clickableName: If True, the widget will be clickable by name.
                - clickableSegment: If True, the widget will be clickable by segment.
                - maxHeight: Maximum height of the widget in pixels.
            - imageSliderData: Dictionary with the following fields:
                - name: Name of the row.
                - preview: Preview image.
                - moreExamples: List of preview images.
            - imageSliderOptions: Dictionary with the following fields:
                - selectable: If True, the widget will be selectable.
                - height: Height of the rows in pixels.

        :returns: Dictionary with widget data.
        :rtype: Dict[str, Any]
        """
        return {
            "content": {
                "maxValue": self._max_value,
                "segments": deepcopy(self._segments),
                "rows": deepcopy(self._rows_data),
            },
            "options": {
                "selectable": self._selectable,
                "collapsable": self._collapsable,
                "clickableName": self._clickable_name,
                "clickableSegment": self._clickable_segment,
                "maxHeight": self._max_height,
            },
            "imageSliderData": deepcopy(self._slider_data),
            "imageSliderOptions": {"selectable": False, "height": self._rows_height},
        }

    def get_json_state(self) -> Dict[str, Any]:
        """Returns dictionary with widget state.

        Dictionary contains the following fields:
            - selectedRows: List of selected rows.
            - clickedItem: Name of the clicked item.

        :returns: Dictionary with widget state.
        :rtype: Dict[str, Any]
        """
        return {"selectedRows": [], "clickedItem": None}

    @property
    def is_selectable(self) -> bool:
        """Returns True if the widget is selectable, False otherwise.

        :returns: True if the widget is selectable, False otherwise.
        :rtype: bool
        """
        self._selectable = DataJson()[self.widget_id]["options"]["selectable"]
        return self._selectable

    @property
    def is_collapsable(self) -> bool:
        """Returns True if the widget is collapsable, False otherwise.

        :returns: True if the widget is collapsable, False otherwise.
        :rtype: bool
        """
        self._collapsable = DataJson()[self.widget_id]["options"]["collapsable"]
        return self._collapsable

    @property
    def is_clickable_name(self) -> bool:
        """Returns True if the widget is clickable by name, False otherwise.

        :returns: True if the widget is clickable by name, False otherwise.
        :rtype: bool
        """
        self._clickable_name = DataJson()[self.widget_id]["options"]["clickableName"]
        return self._clickable_name

    @property
    def is_clickable_segment(self) -> bool:
        """Returns True if the widget is clickable by segment, False otherwise.

        :returns: True if the widget is clickable by segment, False otherwise.
        :rtype: bool
        """
        self._clickable_segment = DataJson()[self.widget_id]["options"]["clickableSegment"]
        return self._clickable_segment

    def get_max_value(self) -> int:
        """Returns maximum value of the widget.

        :returns: Maximum value of the widget.
        :rtype: int
        """
        self._max_value = DataJson()[self.widget_id]["content"]["maxValue"]
        return self._max_value

    def set_max_value(self, value: int) -> None:
        """Sets maximum value of the widget.

        :param value: Maximum value of the widget.
        :type value: int
        """
        self._max_value = value
        DataJson()[self.widget_id]["content"]["maxValue"] = self._max_value
        DataJson().send_changes()

    def set_max_height(self, value: int) -> None:
        """Sets maximum height of the widget.

        :param value: Maximum height of the widget in pixels.
        :type value: int
        :raises TypeError: If max height value is not an integer.
        """
        if not isinstance(value, int):
            raise TypeError("Max height must be an integer.")
        self._max_height = f"{value}px"
        DataJson()[self.widget_id]["options"]["maxHeight"] = self._max_height
        DataJson().send_changes()

    def get_max_height(self) -> int:
        """Returns maximum height of the widget.

        :returns: Maximum height of the widget in pixels.
        :rtype: int
        """
        self._max_height = DataJson()[self.widget_id]["options"]["maxHeight"]
        return int(self._max_height[:-2])

    def add_segments(self, segments: Optional[List[Dict]] = []) -> None:
        """Appends list of segments to the existing segments.

        This method will not overwrite the existing segments, but append to it.
        To overwrite segments, use :meth:`set_segments`.

        :param segments: List of segments to be displayed in the widget.
        :type segments: Optional[List[Dict]]
        :raises ValueError: If segment name or key already exists.
        """
        for curr_segment in segments:
            if curr_segment["name"] in [seg["name"] for seg in self._segments]:
                raise ValueError("Segment name already exists.")
            if curr_segment["key"] in [seg["key"] for seg in self._segments]:
                raise ValueError("Segment key already exists.")
            self._segments.append(curr_segment)

        self.update_data()
        DataJson().send_changes()

    def get_segments(self) -> List[Dict]:
        """Returns list of segments to be displayed in the widget.

        :returns: List of segments to be displayed in the widget.
        :rtype: List[Dict]
        """
        self._segments = DataJson()[self.widget_id]["content"]["segments"]
        return self._segments

    def set_segments(self, segments: Optional[List[Dict]] = []) -> None:
        """Sets list of segments to be displayed in the widget.

        This method will overwrite the existing segments, not append to it.
        To append segments, use :meth:`add_segments`.

        :param segments: List of segments to be displayed in the widget.
        :type segments: Optional[List[Dict]]
        """
        self._segments = segments
        self.update_data()
        DataJson().send_changes()

    def add_rows_data(self, rows_data: Optional[List[Dict]] = []) -> None:
        """Appends list of rows to the existing rows.

        This method will not overwrite the existing rows, but append to it.
        To overwrite rows, use :meth:`set_rows_data`.

        :param rows_data: List of rows to be displayed in the widget.
        :type rows_data: Optional[List[Dict]]
        :raises ValueError: If row data with the same name already exists.
        """
        for curr_rows_data in rows_data:
            if curr_rows_data["name"] in [row_data["name"] for row_data in self._rows_data]:
                raise ValueError("Row data with the same name already exists.")
            self._rows_data.append(curr_rows_data)

        self.update_data()
        DataJson().send_changes()

    def get_rows_data(self) -> List[Dict]:
        """Returns list of rows to be displayed in the widget.

        :returns: List of rows to be displayed in the widget.
        :rtype: List[Dict]
        """
        self._rows_data = DataJson()[self.widget_id]["content"]["rows"]
        return self._rows_data

    def set_rows_data(self, rows_data: Optional[List[Dict]] = []) -> None:
        """Sets list of rows to be displayed in the widget.

        This method will overwrite the existing rows, not append to it.
        To append rows, use :meth:`add_rows_data`.

        :param rows_data: List of rows to be displayed in the widget.
        :type rows_data: Optional[List[Dict]]
        """
        self._rows_data = rows_data
        self.update_data()
        DataJson().send_changes()

    def add_slider_data(self, slider_data: Optional[Dict] = {}) -> None:
        """Appends dictionary of slider data to the existing slider data.

        This method will not overwrite the existing slider data, but append to it.
        To overwrite slider data, use :meth:`set_slider_data`.

        :param slider_data: Dictionary of slider data to be displayed in the widget.
        :type slider_data: Optional[Dict[str, List]]
        :raises ValueError: If row data with the same name does not exist.
        """
        for slider_key, slider_value in slider_data.items():
            if slider_key not in [row_data["name"] for row_data in self._rows_data]:
                raise ValueError("Row data with the same name does not exist.")
            self._slider_data[slider_key] = slider_value

        self.update_data()
        DataJson().send_changes()

    def get_slider_data(self) -> Dict[str, List]:
        """Returns dictionary of slider data to be displayed in the widget.

        :returns: Dictionary of slider data to be displayed in the widget.
        :rtype: Dict[str, List]
        """
        self._slider_data = DataJson()[self.widget_id]["imageSliderData"]
        return self._slider_data

    def set_slider_data(self, slider_data: Optional[Dict[str, List]] = {}):
        """Sets dictionary of slider data to be displayed in the widget.

        This method will overwrite the existing slider data, not append to it.
        To append slider data, use :meth:`add_slider_data`.

        :param slider_data: Dictionary of slider data to be displayed in the widget.
        :type slider_data: Optional[Dict[str, List]]
        :raises ValueError: If row data with the same name does not exist.
        """
        for slider_key in slider_data.keys():
            if slider_key not in [row_data["name"] for row_data in self._rows_data]:
                raise ValueError("Row data with the same name does not exist.")
        self._slider_data = slider_data
        self.update_data()
        DataJson().send_changes()

    def get_selected_rows(self) -> List[str]:
        """Returns list of selected rows.

        :returns: List of selected rows.
        :rtype: List[str]
        """
        return StateJson()[self.widget_id]["selectedRows"]

    def click(self, func: Callable[[str], None]) -> Callable[[], None]:
        """Decorator for the function to be called when the widget is clicked.

        :param func: Function to be called when the widget is clicked.
        :type func: Callable[[str], None]
        :returns: Decorated function.
        :rtype: Callable[[], None]
        """
        route_path = self.get_route_path(ClassBalance.Routes.CLICK)
        server = self._sly_app.get_server()

        self._click_handled = True

        @server.post(route_path)
        def _click():
            res = StateJson()[self.widget_id]["clickedItem"]
            func(res)

        return _click



from __future__ import annotations

from typing import Dict, Literal

from supervisely.nn.benchmark.object_detection.base_vis_metric import DetectionVisMetric
from supervisely.nn.benchmark.visualization.widgets import (
    ChartWidget,
    ContainerWidget,
    MarkdownWidget,
    RadioGroupWidget,
)


class FrequentlyConfused(DetectionVisMetric):
    """Frequently confused class pairs chart for detection reports."""

    MARKDOWN = "frequently_confused"
    MARKDOWN_EMPTY = "frequently_confused_empty"
    CHART = "frequently_confused"
    RADIO_GROUP = "frequently_confused_radio_group"

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.clickable = True
        self.df = self.eval_result.mp.frequently_confused()
        self._keypair_sep: str = "-"
        self.is_empty = self.df.empty
        self.switchable = True

    @property
    def md(self) -> MarkdownWidget:
        text = self.vis_texts.markdown_frequently_confused
        pair = self.df["category_pair"][0]
        prob = self.df["probability"][0]
        text = text.format(
            pair[0],
            pair[1],
            prob.round(2),
            pair[0],
            pair[1],
            (prob * 100).round(),
            pair[0],
            pair[1],
            pair[1],
            pair[0],
        )
        return MarkdownWidget(self.MARKDOWN, "Frequently Confused Classes", text)

    @property
    def chart(self) -> ContainerWidget:
        return ContainerWidget(
            [self.radio_group(), self._get_chart("probability"), self._get_chart("count")],
            self.CHART,
        )

    def radio_group(self) -> RadioGroupWidget:
        return RadioGroupWidget(
            "Probability or Count",
            self.RADIO_GROUP,
            ["probability", "count"],
        )

    def _get_chart(self, switch_key: Literal["probability", "count"]) -> ChartWidget:
        chart = ChartWidget(
            self.CHART,
            self._get_figure(switch_key),
            switch_key=switch_key,
            switchable=self.switchable,
            radiogroup_id=self.RADIO_GROUP,
        )
        chart.set_click_data(
            self.explore_modal_table.id,
            self.get_click_data(),
            chart_click_extra="'getKey': (payload) => `${payload.points[0].x}`, 'keySeparator': '-',",
        )
        return chart

    @property
    def empty_md(self) -> MarkdownWidget:
        text = self.vis_texts.markdown_frequently_confused_empty
        return MarkdownWidget(self.MARKDOWN_EMPTY, "Frequently Confused Classes", text)

    def _get_figure(self, switch_key: Literal["probability", "count"]):  # -> go.Figure:
        if self.is_empty:
            return

        import plotly.graph_objects as go  # pylint: disable=import-error

        # Frequency of confusion as bar chart
        confused_df = self.eval_result.mp.frequently_confused()
        confused_name_pairs = confused_df["category_pair"]
        x_labels = [f"{pair[0]} - {pair[1]}" for pair in confused_name_pairs]
        y_labels = confused_df[switch_key]

        fig = go.Figure()
        fig.add_trace(
            go.Bar(x=x_labels, y=y_labels, marker=dict(color=y_labels, colorscale="Reds"))
        )
        fig.update_layout(
            # title="Frequently confused class pairs",
            xaxis_title="Class Pair",
            yaxis_title=y_labels.name.capitalize(),
            width=1000 if len(x_labels) > 10 else 600,
        )
        fig.update_traces(text=y_labels.round(2))
        fig.update_traces(
            hovertemplate="Class Pair: %{x}<br>"
            + y_labels.name.capitalize()
            + ": %{y:.2f}<extra></extra>"
        )
        return fig

    def get_click_data(self) -> Dict:
        if not self.clickable or self.is_empty:
            return
        res = dict(projectMeta=self.eval_result.pred_project_meta.to_json())

        res["layoutTemplate"] = [None, None, None]
        res["clickData"] = {}

        for keypair, v in self.eval_result.click_data.frequently_confused.items():
            subkey1, subkey2 = keypair
            key = f"{subkey1} {self._keypair_sep} {subkey2}"
            res["clickData"][key] = {}
            res["clickData"][key]["imagesIds"] = []
            res["clickData"][key]["title"] = f"Confused classes: {subkey1} - {subkey2}"

            img_ids = set()
            obj_ids = set()
            for x in v:
                img_ids.add(x["dt_img_id"])
                obj_ids.add(x["dt_obj_id"])

            res["clickData"][key]["imagesIds"] = list(img_ids)
            res["clickData"][key]["filters"] = [
                {"type": "specific_objects", "tagId": None, "value": list(obj_ids)},
            ]

        return res



import os
import numpy as np
from .api_server import start_api_server
from .request_queue import RequestQueue, RequestType
from .incremental_dataset import IncrementalDataset
from .helpers import ClassMap
from .evaluator import LiveEvaluator
import supervisely as sly
from supervisely import logger
from supervisely.nn import TaskType
from datetime import datetime
import signal
import sys
import time
from .checkpoint_utils import resolve_checkpoint, save_state_json
from .artifacts_utils import upload_artifacts
from .loss_plateau_detector import LossPlateauDetector
from pathlib import Path

class Phase:
    """String constants describing the live-training lifecycle phases."""

    READY_TO_START = "ready_to_start"
    WAITING_FOR_SAMPLES = "waiting_for_samples"
    INITIAL_TRAINING = "initial_training"
    TRAINING = "training"

class LiveTraining:
    """Base implementation of an interactive/live training loop driven by requests (start/add sample/predict/status)."""

    from torch import nn  # pylint: disable=import-error
    task_type: str = None  # Should be set in subclass
    framework_name: str = None # Should be set in subclass

    _task2geometries = {
        TaskType.OBJECT_DETECTION: [sly.Rectangle],
        TaskType.INSTANCE_SEGMENTATION: [sly.Bitmap, sly.Polygon],
        TaskType.SEMANTIC_SEGMENTATION: [sly.Bitmap, sly.Polygon],
    }

    def __init__(
            self,
            initial_samples: int = 2,
            filter_classes_by_task: bool = True,
        ):
        """
        :param initial_samples: Min samples before training.
        :type initial_samples: int
        :param filter_classes_by_task: Filter obj classes by task geometry.
        :type filter_classes_by_task: bool
        """
        from torch import nn  # pylint: disable=import-error
        self.initial_samples = initial_samples
        self.filter_classes_by_task = filter_classes_by_task
        if self.task_type is None and self.filter_classes_by_task:
            raise ValueError("task_type must be set in subclass if filter_classes_by_task is set to True")
        if self.framework_name is None:
            raise ValueError("framework_name must be set in subclass")

        self.project_id = sly.env.project_id()
        self.team_id = sly.env.team_id()
        self.task_id = sly.env.task_id(raise_not_found=False)
        self.app = sly.Application()
        self.api = sly.Api()
        self.request_queue = RequestQueue()

        if os.getenv("DEVELOP_AND_DEBUG") and not sly.is_production():
            logger.info(f"🔧 Initializing Develop & Debug application for project {self.project_id}...")
            sly.app.development.supervisely_vpn_network(action="up")
            debug_task = sly.app.development.create_debug_task(self.team_id, port="8000", project_id=self.project_id)
            self.task_id = debug_task['id']
        self._api_thread = start_api_server(self.app, self.request_queue)
        self.phase = Phase.READY_TO_START
        self.iter = 0
        self._loss = None
        self._is_paused = False
        self._should_pause_after_continue = False
        self.initial_iters = 60  # TODO: remove later
        self.project_meta = self._fetch_project_meta(self.project_id)
        self.class_map = self._init_class_map(self.project_meta)
        self.dataset: IncrementalDataset = None
        self.model: nn.Module  = None
        self.loss_plateau_detector = self._init_loss_plateau_detector()
        self.work_dir = 'app_data'
        self.latest_checkpoint_path = f"{self.work_dir}/checkpoints/latest.pth"

        self.checkpoint_mode = os.getenv("modal.state.checkpointMode", "scratch")
        selected_task_id_env = os.getenv("modal.state.selectedExperimentTaskId")
        self.selected_experiment_task_id = int(selected_task_id_env) if selected_task_id_env else None

        self.training_start_time = None
        self._upload_in_progress = False

        self.evaluator = self.init_evaluator()
        self._upload_interval = 7200
        self._last_upload_time = None

        self._inactivity_timeout = 24 * 3600 # 24 hours in seconds
        self._last_activity_time = None

        # from . import live_training_instance
        # live_training_instance = self  # for access from other modules

    @property
    def ready_to_predict(self):
        return self.iter > self.initial_iters

    def status(self):
        return {
            'phase': self.phase,
            'samples_count': len(self.dataset) if self.dataset is not None else 0,
            'waiting_samples': self.initial_samples,
            'task_type': self.task_type,
            'iteration': self.iter,
            'loss': self._loss,
            'training_paused': self._is_paused,
            'ready_to_predict': self.ready_to_predict,
            'initial_iters': self.initial_iters,
            'model_quality': self.evaluator.ema_value if self.evaluator else None
        }

    def run(self):
        self.training_start_time = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        self._add_shutdown_callback()
        self._last_upload_time = time.time()
        self._last_activity_time = time.time()

        work_dir_path = Path(self.work_dir)
        work_dir_path.mkdir(parents=True, exist_ok=True)
        model_meta_path = work_dir_path / "model_meta.json"
        sly.json.dump_json_file(self.project_meta.to_json(), str(model_meta_path))

        try:
            self.phase = Phase.READY_TO_START
            self._wait_for_start()
            if self.checkpoint_mode == 'continue':
                self._run_continue()
            elif self.checkpoint_mode == 'finetune':
                self._run_finetune()
            else:
                self._run_from_scratch()
        except Exception as e:
            if not sly.is_production():
                raise e
            else:
                logger.error(f"Live training failed: {e}", exc_info=True)
                final_checkpoint = self.latest_checkpoint_path
                self.save_checkpoint(final_checkpoint)
                save_state_json(self.state(), final_checkpoint)
                self._upload_artifacts()

    def _run_from_scratch(self):
        self.phase = Phase.WAITING_FOR_SAMPLES
        self._wait_for_initial_samples()
        self.train(checkpoint_path=None)

    def _run_continue(self):
        checkpoint_path, state = self._load_checkpoint()
        self.load_state(state)
        image_ids = state.get('image_ids', [])
        if image_ids:
            self._restore_dataset(image_ids)
        self.train(checkpoint_path=checkpoint_path)

    def _run_finetune(self):
        checkpoint_path, _ = self._load_checkpoint()
        self.phase = Phase.WAITING_FOR_SAMPLES
        self._wait_for_initial_samples()
        self.train(checkpoint_path=checkpoint_path)

    def _wait_for_start(self):
        request = self.request_queue.get()
        while request.type != RequestType.START:
            if request.type == RequestType.STATUS:
                status = self.status()
                request.future.set_result(status)
            else:
                request.future.set_exception(Exception(f"Unexpected request {request.type} while waiting for START"))
            request = self.request_queue.get()
        # When START is received
        status = self.status()
        status['phase'] = Phase.WAITING_FOR_SAMPLES
        request.future.set_result(status)

    def _wait_until_samples_added(
        self,
        samples_needed: int,
        max_wait_time: int = None,
    ):
        sleep_interval = 0.5
        elapsed_time = 0
        samples_before = len(self.dataset)

        while len(self.dataset) - samples_before < samples_needed:
            if max_wait_time is not None and elapsed_time >= max_wait_time:
                raise RuntimeError("Timeout waiting for samples")

            if not self.request_queue.is_empty():
                self._process_pending_requests()

            if self._should_upload_periodically():
                logger.info(f"Periodic upload (interval: {self._upload_interval}s)")
                self._save_and_upload()
                self._last_upload_time = time.time()

            if self._should_stop():
                logger.warning(f"No activity for {self._inactivity_timeout / 3600:.1f} hours")
                self._save_and_upload()
                sys.exit(0)

            time.sleep(sleep_interval)
            elapsed_time += sleep_interval

    def _wait_for_initial_samples(self):
        if len(self.dataset) >= self.initial_samples:
            return

        self.phase = Phase.WAITING_FOR_SAMPLES
        self._is_paused = True

        samples_needed = self.initial_samples - len(self.dataset)
        logger.info(f"Waiting for {samples_needed} initial samples")
        self._wait_until_samples_added(
            samples_needed=samples_needed,
            max_wait_time=3600,
        )

        self._is_paused = False

    def _process_pending_requests(self):
        requests = self.request_queue.get_all()
        if not requests:
            return

        new_samples_added = False

        for request in requests: 
            try:
                if request.type == RequestType.PREDICT:
                    result = self._handle_predict(request.data)
                    request.future.set_result(result)
                    self._last_activity_time = time.time()

                elif request.type == RequestType.ADD_SAMPLE:
                    result = self._handle_add_sample(request.data)
                    request.future.set_result(result)
                    new_samples_added = True
                    self._last_activity_time = time.time()

                elif request.type == RequestType.STATUS:
                    result = self.status()
                    request.future.set_result(result)

            except Exception as e:
                logger.error(f"Error processing request {request.type}: {e}", exc_info=True)
                request.future.set_exception(e)

    def train(self, checkpoint_path: str = None):
        """
        Main training loop. Implement framework-specific training logic here.
        Prepare model config, set hyperparameters and run training.
        Handle phases: initial training, training
        """
        raise NotImplementedError

    def predict(self, model: nn.Module, image_np, image_info) -> list:
        """
        Run inference on a single image and return predictions as a list of sly figures in json format.
        """
        raise NotImplementedError

    def _handle_predict(self, data: dict):
        image_np = data['image']
        image_id = data['image_id']
        image_info = {'id': image_id}   
        model = self.model
        was_training = model.training
        model.eval()
        try:
            objects_raw = self.predict(self.model, image_np=image_np, image_info=image_info)

            if self.evaluator:
                image_shape = image_np.shape[:2]
                self.evaluator.store_prediction(image_id, objects_raw, image_shape)

            return {
                'objects': objects_raw,
                'image_id': image_id,
                'status': self.status(),
            }
        finally:
            # Restore training mode
            if was_training:
                model.train()

    def add_sample(
            self,
            image_id: int,
            image_np: np.ndarray,
            annotation: sly.Annotation,
            image_name: str
        ) -> dict:
        return self.dataset.add_or_update(image_id, image_np, annotation, image_name)

    def _handle_add_sample(self, data: dict):
        ann_json = data['annotation']
        ann_json = self._filter_annotation(ann_json)
        sly_ann = sly.Annotation.from_json(ann_json, self.project_meta)
        image_id = data['image_id']
        self.add_sample(
            image_id=image_id,
            image_np=data['image'],
            annotation=sly_ann,
            image_name=data['image_name']
        )
        if self.evaluator and self.phase!=Phase.WAITING_FOR_SAMPLES:
            result = self.evaluator.evaluate(image_id, sly_ann)
            if result is not None:
                metric_name = self.evaluator.metric_name
                logger.info(
                    f"Image {image_id}: {metric_name}={result['metric_value']:.3f}, "
                    f"EMA={result['ema_value']:.3f}"
                )

        if (len(self.dataset) >= self.initial_samples) and self.phase==Phase.WAITING_FOR_SAMPLES:
            self.phase = Phase.INITIAL_TRAINING

        return {
            'image_id': image_id,
            'status': self.status(),
        }

    def _fetch_project_meta(self, project_id: int) -> sly.ProjectMeta:
        project_meta = self.api.project.get_meta(project_id)
        project_meta = sly.ProjectMeta.from_json(project_meta)
        return project_meta

    def _init_class_map(self, project_meta: sly.ProjectMeta) -> ClassMap:
        obj_classes = list(project_meta.obj_classes)

        if self.task_type == TaskType.SEMANTIC_SEGMENTATION:
            obj_classes.insert(0, sly.ObjClass(name='_background_', geometry_type=sly.Bitmap))

        if self.filter_classes_by_task:
            allowed_geometries = self._task2geometries[self.task_type] + [sly.AnyGeometry]
            obj_classes = [
                obj_class for obj_class in obj_classes
                if obj_class.geometry_type in allowed_geometries
            ]

        return ClassMap(obj_classes)

    def _filter_annotation(self, ann_json: dict) -> dict:
        # Filter objects according to class_map
        # Important: Must be filtered before sly.Annotation.from_json due to static project meta
        allowed_geometries = self._task2geometries[self.task_type]
        allowed_geometries = [geom.geometry_name() for geom in allowed_geometries]
        filtered_objects = []
        for obj in ann_json['objects']:
            sly_id = obj['classId']
            if sly_id in self.class_map.sly_ids and obj['geometryType'] in allowed_geometries:
                filtered_objects.append(obj)
        ann_json['objects'] = filtered_objects
        return ann_json

    def after_train_step(self, loss: float):
        self.iter += 1
        self._loss = loss
        if self._should_pause_after_continue:
            self._is_paused = True
            logger.info("Training was paused. Waiting for 1 new sample before resuming...")
            self._wait_until_samples_added(samples_needed=1, max_wait_time=None)
            self._should_pause_after_continue = False
            logger.info("New sample added. Resuming training...")
            self._is_paused = False
        if self.loss_plateau_detector is not None:
            is_plateau = self.loss_plateau_detector.step(loss, self.iter)
            if is_plateau:
                self._is_paused = True
                self._wait_until_samples_added(
                    samples_needed=1,
                    max_wait_time=None,
                )
                self._is_paused = False
                self.loss_plateau_detector.reset()
                logger.debug(f"Resuming training. Next iteration will be {self.iter + 1}")
        self._process_pending_requests()

    def register_model(self, model: nn.Module):
        self.model = model

    def register_dataset(self, dataset: IncrementalDataset):
        assert hasattr(dataset, 'add_or_update'), "Dataset must implement add_or_update method. Consider inheriting from IncrementalDataset."
        self.dataset = dataset

    def _load_checkpoint(self) -> tuple:
        """Resolve and configure checkpoint based on checkpoint_mode."""
        self._process_pending_requests() 
        checkpoint_path, class_map, state = resolve_checkpoint(
            checkpoint_mode=self.checkpoint_mode,
            selected_experiment_task_id=self.selected_experiment_task_id,
            class_map=self.class_map,
            project_meta=self.project_meta,
            api=self.api,
            team_id=self.team_id,
            work_dir=self.work_dir
        )

        self.class_map = class_map  
        self._process_pending_requests() 
        return checkpoint_path, state

    def state(self):
        state = {
            'phase': self.phase,
            'iter': self.iter,
            'loss': self._loss,
            'clases': [cls.name for cls in self.class_map.obj_classes],
            'image_ids': self.dataset.get_image_ids() if self.dataset else [],
            'dataset_size': len(self.dataset) if self.dataset else 0,
            'is_paused': self._is_paused
        }
        return state

    def load_state(self, state: dict):
        self.phase = state.get('phase', Phase.READY_TO_START)
        self.iter = state.get('iter', 0)
        self._loss = state.get('loss', None)
        self.image_ids = state.get('image_ids', [])
        if state.get('is_paused', False):
            self._should_pause_after_continue = True
        dataset_size = state.get('dataset_size', 0)

    def _restore_dataset(self, image_ids: list):
        if not image_ids:
            return

        logger.info(f"Restoring {len(image_ids)} images from Supervisely...")

        restored_count = 0
        for img_id in image_ids:
            img_info = self.api.image.get_info_by_id(img_id)

            if img_info is None:
                logger.warning(f"Image {img_id} not found, skipping")
                continue

            image_np = self.api.image.download_np(img_id)
            ann_json = self.api.annotation.download_json(img_id)
            ann = sly.Annotation.from_json(ann_json, self.project_meta)

            self.dataset.add_or_update(
                image_id=img_id,
                image_np=image_np,
                annotation=ann,
                image_name=img_info.name
            )

            restored_count += 1

            if restored_count % 10 == 0:
                logger.info(f"Restored {restored_count}/{len(image_ids)}")

        logger.info(f"Restored {restored_count} images")

    def prepare_artifacts(self) -> dict:
        """
        Prepare all artifacts for upload (framework-specific).

        Returns:
            Dict with:
                - checkpoint_path: path to checkpoint file
                - checkpoint_info: dict with {name, iteration, loss}
                - config_path: path to config file
                - logs_dir: path to logs directory or None
                - model_name: model name
                - model_config: model configuration dict
                - loss_history: dict with loss history
        """
        raise NotImplementedError(
            f"{self.__class__.__name__} must implement prepare_artifacts()"
        )

    def _get_session_info(self) -> dict:
        """Collect training session context"""
        return {
            'team_id': self.team_id,
            'task_id': self.task_id,
            'project_id': self.project_id,
            'framework_name': self.framework_name,
            'task_type': self.task_type,
            'class_map': self.class_map,
            'start_time': self.training_start_time,
            'train_size': len(self.dataset) if self.dataset else 0,
            'initial_samples': self.initial_samples
        }

    def _upload_artifacts(self):
        if self._upload_in_progress:
            return

        self._upload_in_progress = True

        try:
            session_info = self._get_session_info()
            artifacts = self.prepare_artifacts()

            report_url = upload_artifacts(
                api=self.api,
                session_info=session_info,
                artifacts=artifacts
            )

            logger.info(f"Report: {report_url}")

        except Exception as e:
            logger.error(f"Upload failed: {e}", exc_info=True)

        finally:
            self._upload_in_progress = False

    def save_checkpoint(self, checkpoint_path: str):
        pass

    def _init_loss_plateau_detector(self):
        loss_plateau_detector = LossPlateauDetector(
            window_size=150,
            threshold=0.002,
            patience=3,
            check_interval=1,
        )
        loss_plateau_detector.register_save_checkpoint_callback(self.save_checkpoint)
        return loss_plateau_detector

    def init_evaluator(self):
        return LiveEvaluator(
            task_type=self.task_type,
            class2idx=self.class_map.class2idx,
            ema_alpha=0.1,
            ignore_index=255,
            score_thr=0.3,
        )

    def _add_shutdown_callback(self):
        """Setup graceful shutdown: save experiment on SIGINT/SIGTERM"""
        self._upload_in_progress = False

        def signal_handler(signum, frame):
            if self._upload_in_progress:
                # Already uploading - force exit on second signal
                signal.signal(signal.SIGINT, lambda s, f: sys.exit(1))
                signal.signal(signal.SIGTERM, lambda s, f: sys.exit(1))
                return

            # Save checkpoint and state before upload
            logger.info("Received shutdown signal, saving checkpoint...")
            self._save_and_upload()
            sys.exit(0)

        signal.signal(signal.SIGINT, signal_handler)
        signal.signal(signal.SIGTERM, signal_handler)

    def _is_timeout_reached(self, last_time: float, timeout: int) -> bool:
        """Check if timeout interval has passed since last_time"""
        if last_time is None:
            return False
        if timeout <= 0:
            return False
        elapsed = time.time() - last_time
        return elapsed >= timeout

    def _should_upload_periodically(self) -> bool:
        """Check if periodic upload should be triggered"""
        return self._is_timeout_reached(self._last_upload_time, self._upload_interval)

    def _should_stop(self) -> bool:
        """Check if training should stop due to user inactivity"""
        return self._is_timeout_reached(self._last_activity_time, self._inactivity_timeout)

    def _save_and_upload(self):
        """Save checkpoint, state, and upload artifacts"""
        logger.info("Saving checkpoint and uploading artifacts...")
        self.save_checkpoint(self.latest_checkpoint_path)
        save_state_json(self.state(), self.latest_checkpoint_path)
        self._upload_artifacts()



# coding: utf-8

import os

from supervisely.task.progress import Progress
from supervisely.sly_logger import logger
from supervisely.task.paths import TaskPaths
from supervisely.metric.metric_base import MetricsBase
from supervisely.project.project import Project, OpenMode
from supervisely.annotation.annotation import Annotation

PROJECT_GT = 'project_gt'
PROJECT_PRED = 'project_pred'


class MetricProjectsApplier:
    """Applies a metric across two projects (ground truth and predictions) with compatible structure."""

    def __init__(self, metric: MetricsBase, config: dict):
        """:param metric: Metric instance (IoUMetric, MAPMetric, etc.).
        :param config: Dict with 'project_gt' and optionally 'project_pred' paths.
        """
        self._metric = metric
        self._confg = config
        self._project_gt = self._get_project_or_die(PROJECT_GT)
        # If a separate project with predictions is not set, use the ground truth project as a source of predictions
        # (with potentially different classes/tags defined by the class mapping).
        self._project_pred = (
            self._get_project_or_die(PROJECT_PRED) if (PROJECT_PRED in self._confg) else self._project_gt)
        self._check_projects_compatible_structure()

    def _get_project_or_die(self, project_field_name):
        project_name = self._confg.get(project_field_name)
        if project_name is None:
            raise RuntimeError('Field "{}" does not exist in config.'.format(project_field_name))
        return Project(os.path.join(TaskPaths.DATA_DIR, project_name), OpenMode.READ)

    def _check_projects_compatible_structure(self):
        if self._project_gt.datasets.keys() != self._project_pred.datasets.keys():  # Keys is sorted - ok
            raise RuntimeError('Projects must contain same datasets.')
        if self._project_gt.total_items != self._project_pred.total_items:
            raise RuntimeError('Projects must contain same number of samples.')
        for ds_gt in self._project_gt.datasets:
            ds_pred = self._project_pred.datasets.get(ds_gt.name)
            for sample_name in ds_gt:
                if not ds_pred.item_exists(sample_name):
                    raise RuntimeError('Projects must contain identically named samples in respective datasets. ' +
                                       'Ground truth project has sample {!r} in dataset {!r}, but prediction project ' +
                                       'does not.'.format(sample_name, ds_gt.name))

        logger.info('Projects structure has been read. Samples: {} per project.'.format(self._project_gt.total_items))

    @property
    def project_gt(self):
        return self._project_gt

    @property
    def project_pred(self):
        return self._project_pred

    def run_evaluation(self):
        progress = Progress('metric evaluation', self._project_gt.total_items)
        for ds_name in self._project_gt.datasets.keys():
            ds_gt = self._project_gt.datasets.get(ds_name)
            ds_pred = self._project_pred.datasets.get(ds_name)

            for sample_name in ds_gt:
                try:
                    ann_gt = Annotation.load_json_file(ds_gt.get_ann_path(sample_name), self._project_gt.meta)
                    ann_pred = Annotation.load_json_file(ds_pred.get_ann_path(sample_name), self._project_pred.meta)
                    self._metric.add_pair(ann_gt, ann_pred)
                except ValueError as e:
                    logger.warning('An error has occured ({}). Sample "{}" in dataset "{}" will be skipped'
                                   .format(str(e), sample_name, ds_gt.name))
                progress.iter_done_report()



from typing import Dict, List

from supervisely.nn.benchmark.visualization.widgets.widget import BaseWidget


class RadioGroupWidget(BaseWidget):
    """Benchmark report widget that groups switchable content; one key active at a time."""

    def __init__(
        self,
        name: str,
        radio_group: str,
        switch_keys: List[str],
        default_key: str = None,
    ) -> None:
        """:param name: Widget name.
        :param radio_group: Group ID for state.
        :param switch_keys: Keys for radio options.
        :param default_key: Default selected key.
        """
        super().__init__(name)
        self.radio_group = radio_group
        self.switch_keys = switch_keys
        self.default_key = default_key or switch_keys[0]

    def save_data(self, basepath: str) -> None:
        return

    def get_state(self) -> Dict:
        return {self.radio_group: self.default_key}

    def to_html(self) -> str:
        res = ""
        for switch_key in self.switch_keys:
            res += f"""
            <el-radio v-model="state.{self.radio_group}" label="{switch_key}" style="margin-top: 10px;">
            {switch_key}
            </el-radio>
            """
        return res



import re
import traceback
from functools import wraps
from json import JSONDecodeError, loads
from shutil import ReadError
from tarfile import ReadError as TarReadError
from typing import Callable, Dict, List, Optional, Union

from requests.exceptions import HTTPError, RetryError
from rich.console import Console

from supervisely import is_community, is_development
from supervisely.app import DialogWindowError
from supervisely.sly_logger import EventType, logger

# TODO: Add correct doc link.
# DOC_URL = "https://docs.supervisely.com/errors/"


class HandleException:
    """
    Base wrapper for an exception enriched with a user-facing error code, title and message.

    Used by :class:`ErrorHandler` to present consistent errors in UI (DialogWindowError) and logs.
    """
    def __init__(
        self,
        exception: Exception,
        stack: List[traceback.FrameSummary] = None,
        **kwargs,
    ):
        """:param exception: Original exception.
        :param stack: Optional traceback frame list. Inferred from exception if None.
        :param kwargs: code, title, message for user-facing error.
        """
        self.exception = exception
        self.stack = stack or read_stack_from_exception(self.exception)
        self.code = kwargs.get("code")
        self.title = kwargs.get("title")

        # TODO: Add error code to the title.
        # self.title += f" (SLYE{self.code})"

        self.message = kwargs.get("message")

        # TODO: Add doc link to the message.
        # error_url = DOC_URL + str(self.code)
        # self.message += (
        #     f"<br>Check out the <a href='{error_url}'>documentation</a> "
        #     "for solutions and troubleshooting."
        # )

        if is_development():
            self.dev_logging()

    def dev_logging(self):
        # * Option for logging the exception with sly logger.
        # * Due to JSON formatting it's not convinient to read the log.
        # tracelog = " | ".join(traceback.format_list(self.stack))
        # logger.error(
        #     "Detailed info about exception."
        #     f"Exception type: {type(self.exception)}. "
        #     f"Exception value: {self.exception}. "
        #     f"Error code: {self.code}. "
        #     f"Error title: {self.title}. "
        #     f"Error message: {self.message} "
        #     f"Traceback: {tracelog}. "
        # )

        # * Printing the exception to the console line by line.
        console = Console()

        console.print(self.title, style="bold red")
        console.print(self.message, style="bold red")
        # TODO: Uncomment code line when error codes will be added.
        # console.print(f"Error code: {self.code}.", style="bold orange")
        console.print(f"{self.exception.__class__.__name__}: {self.exception}")

        # console.print("Traceback (most recent call last):", style="bold red")
        # for i, trace in enumerate(traceback.format_list(self.stack)):
        #     console.print(f"{i + 1}. {trace}")

    def raise_error(self, has_ui: bool = True):
        if has_ui:
            raise DialogWindowError(self.title, self.message)
        else:
            raise self.exception.__class__(self.get_message_for_exception()) from self.exception

    def log_error_for_agent(self, main_name: str):
        logger.critical(
            self.get_message_for_exception(),
            exc_info=True,
            extra={
                "main_name": main_name,
                "event_type": EventType.TASK_CRASHED,
                "exc_str": self.message,
            },
        )

    def get_message_for_modal_window(self):
        return f"{self.title}. \n{self.message}"

    def get_message_for_exception(self):
        return f"{self.title}. \n{self.message}"


class ErrorHandler:
    """
    Namespaced collection of typed error wrappers used to convert raw exceptions into user-facing messages.
    """
    class APP:
        """Errors originating from app/runtime logic (UI, files, archives, etc.)."""
        class UnsupportedShapes(HandleException):
            """Raised when project classes contain unsupported geometry shapes."""
            def __init__(self, exception: Exception, stack: List[traceback.FrameSummary] = None):
                """See :class:`~supervisely.io.exception_handlers.HandleException` for params."""
                self.code = 1001
                self.title = "Unsupported class shapes"
                self.message = exception.args[0]

                super().__init__(
                    exception,
                    stack,
                    code=self.code,
                    title=self.title,
                    message=self.message,
                )

        class UnsupportedArchiveFormat(HandleException):
            """Raised when an input archive has an unsupported format."""
            def __init__(self, exception: Exception, stack: List[traceback.FrameSummary] = None):
                """See :class:`~supervisely.io.exception_handlers.HandleException` for params."""
                self.code = 1002
                self.title = "Unsupported archive format"
                self.message = "The given archive format is not supported."

                super().__init__(
                    exception,
                    stack,
                    code=self.code,
                    title=self.title,
                    message=self.message,
                )

        class UnicodeDecodeError(HandleException):
            """Raised when reading a text file fails due to encoding issues."""
            def __init__(self, exception: Exception, stack: List[traceback.FrameSummary] = None):
                """See :class:`~supervisely.io.exception_handlers.HandleException` for params."""
                self.code = 1003
                self.title = "Unicode decode error"
                self.message = (
                    "The given file contains non-unicode characters. "
                    "Please, check the file and try again."
                )

                super().__init__(
                    exception,
                    stack,
                    code=self.code,
                    title=self.title,
                    message=self.message,
                )

        class CallUndeployedModelError(HandleException):
            """Raised when an inference request is made to a model that is not deployed/serving."""
            def __init__(
                self, exception: Exception, stack: List[traceback.FrameSummary] = None, **kwargs
            ):
                """See :class:`~supervisely.io.exception_handlers.HandleException` for params."""
                self.code = 1004
                self.title = "Call undeployed model error"
                self.message = str(exception.args[0])
                super().__init__(
                    exception,
                    stack,
                    code=self.code,
                    title=self.title,
                    message=self.message,
                )

        class FailedToReadArchive(HandleException):
            """Raised when an archive cannot be read (corrupted or unsupported content)."""
            def __init__(self, exception: Exception, stack: List[traceback.FrameSummary] = None):
                """See :class:`~supervisely.io.exception_handlers.HandleException` for params."""
                self.code = 1005
                self.title = "Failed to read archive"
                self.message = (
                    "Error occurred while reading the archive. "
                    "Please, check the archive format, size and it is not corrupted."
                )

                super().__init__(
                    exception,
                    stack,
                    code=self.code,
                    title=self.title,
                    message=self.message,
                )

        class FailedToUnpackArchive(HandleException):
            """Raised when an archive cannot be unpacked (corrupted or unsupported content)."""
            def __init__(self, exception: Exception, stack: List[traceback.FrameSummary] = None):
                """See :class:`~supervisely.io.exception_handlers.HandleException` for params."""
                self.code = 1006
                self.title = "Failed to unpack archive"
                self.message = (
                    "Error occurred while unpacking the archive. "
                    "Please, check the archive format, size and it is not corrupted."
                )

                super().__init__(
                    exception,
                    stack,
                    code=self.code,
                    title=self.title,
                    message=self.message,
                )

    class API:
        """Errors originating from Supervisely API requests and server-side constraints."""
        class TeamFilesFileNotFound(HandleException):
            """Raised when a requested Team Files path does not exist."""
            def __init__(self, exception: Exception, stack: List[traceback.FrameSummary] = None):
                """See :class:`~supervisely.io.exception_handlers.HandleException` for params."""
                self.code = 2001
                self.title = "Requested file was not found in Team Files"
                self.message = (
                    "The requested file doesn't exist in Team Files in current Team. "
                    "Please, ensure that the file exists and you're working in the correct Team."
                )

                super().__init__(
                    exception,
                    stack,
                    code=self.code,
                    title=self.title,
                    message=self.message,
                )

        class TaskSendRequestError(HandleException):
            """Raised when the app cannot send a request to a running task/session."""
            def __init__(self, exception: Exception, stack: List[traceback.FrameSummary] = None):
                """See :class:`~supervisely.io.exception_handlers.HandleException` for params."""
                self.code = 2002
                self.title = "Task send request error. Check the logs for more information."
                self.message = (
                    "The application has encountered an error while sending a request to the task. "
                    "It may be caused by connection issues. "
                    "Please, check your internet connection, the task is running and input parameters. "
                    "If you are using a model in other session, please, check that the model is serving and it's logs."
                )

                super().__init__(
                    exception,
                    stack,
                    code=self.code,
                    title=self.title,
                    message=self.message,
                )

        class FileSizeTooLarge(HandleException):
            """Raised when uploading a file exceeds the plan/file size limits."""
            def __init__(self, exception: Exception, stack: List[traceback.FrameSummary] = None):
                """See :class:`~supervisely.io.exception_handlers.HandleException` for params."""
                self.code = 2003
                self.title = "File size limit exceeded"
                self.message = "The given file size is too large "
                if is_community():
                    self.message += "for Community Edition (more than 10 GB)."
                else:
                    self.message += "(more than 100 GB)."

                super().__init__(
                    exception,
                    stack,
                    code=self.code,
                    title=self.title,
                    message=self.message,
                )

        class ImageFilesSizeTooLarge(HandleException):
            """Raised when uploading an image exceeds size limits."""
            def __init__(self, exception: Exception, stack: List[traceback.FrameSummary] = None):
                """See :class:`~supervisely.io.exception_handlers.HandleException` for params."""
                self.code = 2004
                self.title = "Image files size limit exceeded"
                self.message = (
                    "The given image file size is too large (more than 1 GB) for Community Edition."
                )

                super().__init__(
                    exception,
                    stack,
                    code=self.code,
                    title=self.title,
                    message=self.message,
                )

        class VideoFilesSizeTooLarge(HandleException):
            """Raised when uploading a video exceeds size limits."""
            def __init__(self, exception: Exception, stack: List[traceback.FrameSummary] = None):
                """See :class:`~supervisely.io.exception_handlers.HandleException` for params."""
                self.code = 2005
                self.title = "Video files size limit exceeded"
                self.message = "The given video file size is too large (more than 300 MB) for Community Edition."

                super().__init__(
                    exception,
                    stack,
                    code=self.code,
                    title=self.title,
                    message=self.message,
                )

        class VolumeFilesSizeTooLarge(HandleException):
            """Raised when uploading a volume exceeds size limits."""
            def __init__(self, exception: Exception, stack: List[traceback.FrameSummary] = None):
                """See :class:`~supervisely.io.exception_handlers.HandleException` for params."""
                self.code = 2006
                self.title = "Volume files size limit exceeded"
                self.message = "The given volume file size is too large (more than 150 MB) for Community Edition."

                super().__init__(
                    exception,
                    stack,
                    code=self.code,
                    title=self.title,
                    message=self.message,
                )

        class OutOfMemory(HandleException):
            """Raised when an operation fails due to OOM (CPU or GPU) on the agent machine."""
            def __init__(self, exception: Exception, stack: List[traceback.FrameSummary] = None):
                """See :class:`~supervisely.io.exception_handlers.HandleException` for params."""
                device_prefix = (
                    "GPU memory"
                    if any([s in exception.args[0] for s in ["CUDA", "cuda", "GPU"]])
                    else "memory"
                )
                self.code = 2007
                self.title = f"Out of {device_prefix} on the computer where the agent is deployed"
                self.message = "Please, check your agent's memory usage, reduce batch size or use a device with more memory capacity."

                super().__init__(
                    exception,
                    stack,
                    code=self.code,
                    title=self.title,
                    message=self.message,
                )

        class DockerRuntimeError(HandleException):
            """Raised when agent-side Docker runtime is unavailable or malfunctioning."""
            def __init__(self, exception: Exception, stack: List[traceback.FrameSummary] = None):
                """See :class:`~supervisely.io.exception_handlers.HandleException` for params."""
                self.code = 2008
                self.title = "Docker runtime error"
                self.message = (
                    "The agent has encountered a Docker runtime error. "
                    "Please, check that docker is installed and running."
                )

                super().__init__(
                    exception,
                    stack,
                    code=self.code,
                    title=self.title,
                    message=self.message,
                )

        class AppSetFieldError(HandleException):
            """Raised when updating app/task fields via API fails (often due to connectivity)."""
            def __init__(self, exception: Exception, stack: List[traceback.FrameSummary] = None):
                """See :class:`~supervisely.io.exception_handlers.HandleException` for params."""
                self.code = 2009
                self.title = "App set field error. Check the logs for more information."
                self.message = (
                    "The application has encountered an error while sending a request. "
                    "It may be caused by connection issues. "
                    "Please, check your internet connection, the app is running and input parameters. "
                    "If you are using a model in other session, please, check that the model is serving and it's logs."
                )

                super().__init__(
                    exception,
                    stack,
                    code=self.code,
                    title=self.title,
                    message=self.message,
                )

        class TeamFilesDirectoryDownloadError(HandleException):
            """Raised when downloading a Team Files directory fails."""
            def __init__(self, exception: Exception, stack: List[traceback.FrameSummary] = None):
                """See :class:`~supervisely.io.exception_handlers.HandleException` for params."""
                self.code = 2010
                self.title = "Team files directory download error"
                self.message = "Make sure that the directory exists in the team files, the files are not corrupted, and try again."

                super().__init__(
                    exception,
                    stack,
                    code=self.code,
                    title=self.title,
                    message=self.message,
                )

        class PointcloudsUploadError(HandleException):
            """Raised when point cloud upload API fails."""
            def __init__(self, exception: Exception, stack: List[traceback.FrameSummary] = None):
                """See :class:`~supervisely.io.exception_handlers.HandleException` for params."""
                self.code = 2011
                self.title = "Pointclouds uploading error"
                self.message = exception.args[0]

                super().__init__(
                    exception,
                    stack,
                    code=self.code,
                    title=self.title,
                    message=self.message,
                )

        class AnnotationUploadError(HandleException):
            """Raised when annotation upload API fails."""
            def __init__(self, exception: Exception, stack: List[traceback.FrameSummary] = None):
                """See :class:`~supervisely.io.exception_handlers.HandleException` for params."""
                self.code = 2012
                self.title = "Annotation uploading error"
                self.message = exception.args[0]

                super().__init__(
                    exception,
                    stack,
                    code=self.code,
                    title=self.title,
                    message=self.message,
                )

        class AnnotationNotFound(HandleException):
            """Raised when annotation file(s) are not found for upload/download."""
            def __init__(self, exception: Exception, stack: List[traceback.FrameSummary] = None):
                """See :class:`~supervisely.io.exception_handlers.HandleException` for params."""
                self.code = 2013
                self.title = "Annotation not found"
                self.message = "Please, check that the annotation(s) exists by the given path."

                super().__init__(
                    exception,
                    stack,
                    code=self.code,
                    title=self.title,
                    message=self.message,
                )

        class ServerOverload(HandleException):
            """Raised for transient server overload conditions; suggests retry later."""
            def __init__(self, exception: Exception, stack: List[traceback.FrameSummary] = None):
                """See :class:`~supervisely.io.exception_handlers.HandleException` for params."""
                self.code = 2014
                self.title = "High load on the server"
                self.message = "Sorry, the server is overloaded. Please, try again later."

                super().__init__(
                    exception,
                    stack,
                    code=self.code,
                    title=self.title,
                    message=self.message,
                )

        class ProjectNotFound(HandleException):
            """Raised when a project cannot be accessed (missing/archived/permissions)."""
            def __init__(self, exception: Exception, stack: List[traceback.FrameSummary] = None):
                """See :class:`~supervisely.io.exception_handlers.HandleException` for params."""
                self.code = 2015
                self.title = "Project not found"
                self.message = "Please, check that the project exists, not archived and you have enough permissions to access it"

                super().__init__(
                    exception,
                    stack,
                    code=self.code,
                    title=self.title,
                    message=self.message,
                )

        class DatasetNotFound(HandleException):
            """Raised when a dataset cannot be accessed (missing/archived/permissions)."""
            def __init__(self, exception: Exception, stack: List[traceback.FrameSummary] = None):
                """See :class:`~supervisely.io.exception_handlers.HandleException` for params."""
                self.code = 2016
                self.title = "Dataset not found"
                self.message = "Please, check that the dataset exists, not archived and you have enough permissions to access it."

                super().__init__(
                    exception,
                    stack,
                    code=self.code,
                    title=self.title,
                    message=self.message,
                )

        class TaskFinished(HandleException):
            """Raised when attempting to operate on a task that has already finished."""
            def __init__(self, exception: Exception, stack: List[traceback.FrameSummary] = None):
                """See :class:`~supervisely.io.exception_handlers.HandleException` for params."""
                self.code = 2017
                self.title = "Task finished"
                self.message = "The task is already finished. Please, check the task's logs for more information."

                super().__init__(
                    exception,
                    stack,
                    code=self.code,
                    title=self.title,
                    message=self.message,
                )

        class PaymentRequired(HandleException):
            """Raised when plan usage limits are reached and an upgrade is required."""
            def __init__(self, exception: Exception, stack: List[traceback.FrameSummary] = None):
                """See :class:`~supervisely.io.exception_handlers.HandleException` for params."""
                self.code = 2018
                self.title = "Usage limits reached"
                self.message = "Your plan usage exceeded. Please, upgrade plan to continue: https://supervisely.com/pricing/"

                super().__init__(
                    exception,
                    stack,
                    code=self.code,
                    title=self.title,
                    message=self.message,
                )

        class FreePlanImagesUploadLinksError(HandleException):
            """Raised when trying to upload remote images on a non-PRO plan."""
            def __init__(self, exception: Exception, stack: List[traceback.FrameSummary] = None):
                """See :class:`~supervisely.io.exception_handlers.HandleException` for params."""
                self.code = 2019
                self.title = "Upload remote images is only available for PRO teams"
                self.message = (
                    "Please, upgrade plan to upload remote images: https://supervisely.com/pricing/"
                )

                super().__init__(
                    exception,
                    stack,
                    code=self.code,
                    title=self.title,
                    message=self.message,
                )

        class FreePlanVideosUploadsLinkError(HandleException):
            """Raised when trying to upload remote videos on a non-PRO plan."""
            def __init__(self, exception: Exception, stack: List[traceback.FrameSummary] = None):
                """See :class:`~supervisely.io.exception_handlers.HandleException` for params."""
                self.code = 2020
                self.title = "Upload remote videos is only available for PRO teams"
                self.message = (
                    "Please, upgrade plan to upload remote videos: https://supervisely.com/pricing/"
                )

                super().__init__(
                    exception,
                    stack,
                    code=self.code,
                    title=self.title,
                    message=self.message,
                )

    class SDK:
        """Errors produced by SDK-side parsing, conversion and local project structure checks."""
        class ProjectStructureError(HandleException):
            """Raised when a local project structure is invalid or incomplete."""
            def __init__(self, exception: Exception, stack: List[traceback.FrameSummary] = None):
                """See :class:`~supervisely.io.exception_handlers.HandleException` for params."""
                self.code = 3001
                self.title = "Project structure error"
                self.message = "Please, check the project structure and try again."

                super().__init__(
                    exception,
                    stack,
                    code=self.code,
                    title=self.title,
                    message=self.message,
                )

        class ConversionNotImplemented(HandleException):
            """Raised when a requested geometry conversion is not implemented."""
            def __init__(self, exception: Exception, stack: List[traceback.FrameSummary] = None):
                """See :class:`~supervisely.io.exception_handlers.HandleException` for params."""
                self.code = 3002
                self.title = "Not implemented error"
                self.message = (
                    "Conversion is not implemented between the given object types. "
                    "Please, check the geometry type of the objects."
                )

                super().__init__(
                    exception,
                    stack,
                    code=self.code,
                    title=self.title,
                    message=self.message,
                )

        class JsonAnnotationReadError(HandleException):
            """Raised when reading an annotation JSON fails (invalid JSON or schema)."""
            def __init__(self, exception: Exception, stack: List[traceback.FrameSummary] = None):
                """See :class:`~supervisely.io.exception_handlers.HandleException` for params."""
                self.code = 3003
                self.title = "JSON annotation read error"
                self.message = (
                    "Please, check that the file has the correct Supervisely JSON format."
                )

                super().__init__(
                    exception,
                    stack,
                    code=self.code,
                    title=self.title,
                    message=self.message,
                )

        class LabelFromJsonFailed(HandleException):
            """Raised when deserializing a Label from JSON fails."""
            def __init__(self, exception: Exception, stack: List[traceback.FrameSummary] = None):
                """See :class:`~supervisely.io.exception_handlers.HandleException` for params."""
                self.code = 3004
                self.title = "Label deserialize error"
                self.message = exception.args[0]

                super().__init__(
                    exception,
                    stack,
                    code=self.code,
                    title=self.title,
                    message=self.message,
                )

        class FileNameTooLong(HandleException):
            """Raised when file system operations fail due to excessively long file names."""
            def __init__(self, exception: Exception, stack: List[traceback.FrameSummary] = None):
                """See :class:`~supervisely.io.exception_handlers.HandleException` for params."""
                self.code = 3005
                self.title = "File name too long"
                self.message = f"Please, check name of items: {exception.args[0]}"

                super().__init__(
                    exception,
                    stack,
                    code=self.code,
                    title=self.title,
                    message=self.message,
                )

        class AnnotationFromJsonFailed(HandleException):
            """Raised when deserializing an Annotation from JSON fails."""
            def __init__(self, exception: Exception, stack: List[traceback.FrameSummary] = None):
                """See :class:`~supervisely.io.exception_handlers.HandleException` for params."""
                self.code = 3005
                self.title = "Annotation deserialize error"
                self.message = exception.args[0]

                super().__init__(
                    exception,
                    stack,
                    code=self.code,
                    title=self.title,
                    message=self.message,
                )

    class AgentDocker:
        """Errors related to Docker operations performed by an Agent."""
        @classmethod
        def parse_docker_exception(cls, docker_exception: Exception) -> str:
            default_msg = str(docker_exception)
            if isinstance(docker_exception, HTTPError):
                try:
                    json_text = docker_exception.response.text
                    info = loads(json_text)
                    return info.get("message", default_msg)
                except JSONDecodeError:
                    pass
            return default_msg

        class ImageNotFound(HandleException):
            """Raised when a required Docker image cannot be found/pulled by the agent."""
            def __init__(self, exception: Exception, stack: List[traceback.FrameSummary] = None):
                """See :class:`~supervisely.io.exception_handlers.HandleException` for params."""
                self.code = 4001
                self.title = "Docker image not found"
                self.message = ErrorHandler.AgentDocker.parse_docker_exception(exception)

                super().__init__(
                    exception,
                    stack,
                    code=self.code,
                    title=self.title,
                    message=self.message,
                )

        class NetworkNotFound(HandleException):
            """Raised when a required Docker network is missing or was deleted."""
            def __init__(self, exception: Exception, stack: List[traceback.FrameSummary] = None):
                """See :class:`~supervisely.io.exception_handlers.HandleException` for params."""
                self.code = 4002
                self.title = "Docker image not found"
                self.message = ErrorHandler.AgentDocker.parse_docker_exception(exception)

                connect_url = (
                    "https://developer.supervisely.com/getting-started/connect-your-computer"
                )
                additional = (
                    "Probably your network was deleted. "
                    "Try to restart docker with `systemctl restart docker` "
                    f"or reinstall your Agent using <a href='{connect_url}'>instructions</a>"
                )
                self.message = "{exc_msg}. {add}".format(exc_msg=self.message, add=additional)

                super().__init__(
                    exception,
                    stack,
                    code=self.code,
                    title=self.title,
                    message=self.message,
                )

    class Agent:
        """Errors caused by Agent runtime issues (auth, memory limits, etc.)."""
        class AgentError(HandleException):
            """Raised for general agent-side failures."""
            def __init__(self, exception: Exception, stack: List[traceback.FrameSummary] = None):
                """See :class:`~supervisely.io.exception_handlers.HandleException` for params."""
                self.code = 5001
                self.title = "Agent issue"
                self.message = (
                    "The agent has encountered an error. "
                    "Please, check the agent's logs for more information."
                )

                super().__init__(
                    exception,
                    stack,
                    code=self.code,
                    title=self.title,
                    message=self.message,
                )

        class MemoryExceeded(HandleException):
            """Raised when the task process is killed due to high memory usage (e.g. exit 137)."""
            def __init__(self, exception: Exception, stack: List[traceback.FrameSummary] = None):
                """See :class:`~supervisely.io.exception_handlers.HandleException` for params."""
                self.code = 5002
                self.title = "Process terminated due to high memory usage"
                self.message = (
                    "Memory usage exceeded the limit. "
                    "Please, check the agent's memory usage, or use a device with more memory capacity."
                )

                super().__init__(
                    exception,
                    stack,
                    code=self.code,
                    title=self.title,
                    message=self.message,
                )


ERROR_PATTERNS = {
    AttributeError: {r".*api\.file\.get_info_by_path.*": ErrorHandler.API.TeamFilesFileNotFound},
    HTTPError: {
        r".*api\.task\.send_request.*": ErrorHandler.API.TaskSendRequestError,
        r".*api\.app\.set_field.*": ErrorHandler.API.AppSetFieldError,
        r".*file-storage\.bulk\.upload.*File too large.*": ErrorHandler.API.FileSizeTooLarge,
        r".*images\.bulk\.upload.*FileSize.*\"sizeLimit\":1073741824.*": ErrorHandler.API.ImageFilesSizeTooLarge,
        r".*videos\.bulk\.upload.*FileSize.*sizeLimit\":314572800.*": ErrorHandler.API.VideoFilesSizeTooLarge,
        r".*images\.bulk\.upload.*FileSize.*\"sizeLimit\":157286400.*": ErrorHandler.API.VolumeFilesSizeTooLarge,
        r".*images\.bulk\.upload.*FileSize.*\"sizeLimit\":\"25mb\".*": ErrorHandler.API.VolumeFilesSizeTooLarge,
        r".*Dataset with datasetId.*is either archived, doesn't exist or you don't have enough permissions to access.*": ErrorHandler.API.DatasetNotFound,
        r".*Project with projectId.*is either archived, doesn't exist or you don't have enough permissions to access.*": ErrorHandler.API.ProjectNotFound,
        r".*api\.task\.set_field.*": ErrorHandler.API.AppSetFieldError,
        r".*Unauthorized for url.*": ErrorHandler.Agent.AgentError,
        r".*Operation\ is\ canceled\ because\ task\ with\ id.*already\ finished.*": ErrorHandler.API.TaskFinished,
        r".*Payment\ Required.*": ErrorHandler.API.PaymentRequired,
        r".*images\.bulk\.add.*only\ available\ for\ PRO\ teams": ErrorHandler.API.FreePlanImagesUploadLinksError,
        r".*videos\.bulk\.add.*only\ available\ for\ PRO\ teams": ErrorHandler.API.FreePlanVideosUploadsLinkError,
    },
    RuntimeError: {
        r".*Label\.from_json.*": ErrorHandler.SDK.LabelFromJsonFailed,
        r".*Annotation\.from_json.*": ErrorHandler.SDK.AnnotationFromJsonFailed,
        r".*out\sof\smemory.*": ErrorHandler.API.OutOfMemory,
        r".*cuda runtime error.*out of memory.*": ErrorHandler.API.OutOfMemory,
        r".*CUDA error.*an illegal memory access was encountered.*": ErrorHandler.API.OutOfMemory,
        r"The\ model\ has\ not\ yet\ been\ deployed.*": ErrorHandler.APP.CallUndeployedModelError,
        r".*Task\ container\ finished\ with\ non-zero\ status: 137.*": ErrorHandler.Agent.MemoryExceeded,
    },
    FileNotFoundError: {
        r".*api\.annotation\.upload_path.*": ErrorHandler.API.AnnotationNotFound,
        r".*api\.annotation\.upload_paths.*": ErrorHandler.API.AnnotationNotFound,
        r".*read_single_project.*": ErrorHandler.SDK.ProjectStructureError,
    },
    JSONDecodeError: {
        r".*sly\.json\.load_json_file.*": ErrorHandler.SDK.JsonAnnotationReadError,
        r".*api\.annotation\.upload_path.*": ErrorHandler.SDK.JsonAnnotationReadError,
        r".*api\.annotation\.upload_paths.*": ErrorHandler.SDK.JsonAnnotationReadError,
    },
    UnicodeDecodeError: {
        r".*codec can't decode byte.*": ErrorHandler.APP.UnicodeDecodeError,
    },
    NotImplementedError: {
        r".*geometry\.convert.*": ErrorHandler.SDK.ConversionNotImplemented,
    },
    ValueError: {
        r".*obj_class\.geometry_type\.from_json.*": ErrorHandler.SDK.LabelFromJsonFailed,
    },
    IndexError: {
        r".*obj_class\.geometry_type\.from_json.*": ErrorHandler.SDK.LabelFromJsonFailed,
    },
    ReadError: {
        r".*unexpected end of data.*": ErrorHandler.APP.FailedToReadArchive,
        r".*Unknown archive format.*": ErrorHandler.APP.UnsupportedArchiveFormat,
        r".*shutil\.unpack_archive.*": ErrorHandler.APP.FailedToUnpackArchive,
        r".*api\.file\.download_directory.*": ErrorHandler.API.TeamFilesDirectoryDownloadError,
    },
    TarReadError: {
        r".*unexpected end of data.*": ErrorHandler.APP.FailedToReadArchive,
        r".*extractall.*": ErrorHandler.APP.FailedToUnpackArchive,
    },
    KeyError: {
        r".*api\.pointcloud\.upload_paths.*": ErrorHandler.API.PointcloudsUploadError,
        r".*api\.pointcloud\.upload_project.*": ErrorHandler.SDK.ProjectStructureError,
        r".*api\.image\.download_bytes.*": ErrorHandler.API.ServerOverload,
        r".*api\.video\.frame\.download_np.*": ErrorHandler.API.ServerOverload,
        r".*api\.image\.download_bytes.*": ErrorHandler.API.ServerOverload,
    },
    TypeError: {
        r".*obj_class\.geometry_type\.from_json.*": ErrorHandler.SDK.LabelFromJsonFailed,
    },
    RetryError: {
        r".*api\.annotation\.upload_paths.*": ErrorHandler.API.AnnotationUploadError,
        r".*api\.task\.set_field.*": ErrorHandler.API.AppSetFieldError,
        r".*api\.app\.set_field.*": ErrorHandler.API.AppSetFieldError,
        r".*api\.task\.send_request.*": ErrorHandler.API.TaskSendRequestError,
    },
    OSError: {
        r".*Errno 36.*File name too long.*": ErrorHandler.SDK.FileNameTooLong,
    },
}

try:
    from docker.errors import ImageNotFound, NotFound

    docker_patterns = {
        ImageNotFound: {
            r".*pull\ access\ denied\ for\ (.*)?,\ repository does not exist.*": ErrorHandler.AgentDocker.ImageNotFound
        },
        NotFound: {
            r".*network\ supervisely-net-(.*)?\ not\ found.*": ErrorHandler.AgentDocker.NetworkNotFound
        },
    }
except ModuleNotFoundError:
    docker_patterns = {}

ERROR_PATTERNS.update(docker_patterns)


def handle_exception(exception: Exception) -> Union[HandleException, None]:
    """Function for handling exceptions, using the stack trace and patterns for known errors.
    Returns an instance of the ErrorHandler class if the pattern is found, otherwise returns None.

    :param exception: Exception to be handled.
    :type exception: Exception
    :returns: Instance of the :class:`~supervisely.io.exception_handlers.ErrorHandler` class or None.
    :rtype: Union[:class:`~supervisely.io.exception_handlers.ErrorHandler`, None]

    :Usage Example:

        .. code-block:: python

            import supervisely as sly

            try:
                # Code that may raise an exception.
                raise Exception("Test exception")
            except Exception as e:
                exception_handler = sly.handle_exception(e)
                if exception_handler:
                    # You may raise the exception using the raise_error() method.
                    # Or use other approaches for handling the exception (e.g. logging, printing to console, etc.)
                    exception_handler.raise_error()
                else:
                    # If the pattern is not found, the exception is raised as usual.
                    raise
    """
    # Extracting the stack trace.
    stack = read_stack_from_exception(exception)

    # Retrieving the patterns for the given exception type.
    patterns = ERROR_PATTERNS.get(type(exception))
    if not patterns:
        return

    # Looping through the stack trace from the bottom up to find matching pattern with specified Exception type.
    for pattern, handler in patterns.items():
        for frame in stack[::-1]:
            if re.match(pattern, frame.line):
                return handler(exception, stack)
        arg = next(iter(exception.args), None)
        if isinstance(arg, str) and re.match(pattern, arg):
            return handler(exception, stack)
        if isinstance(exception, HTTPError):
            msg = exception.response.text
            if re.match(pattern, msg):
                return handler(exception, stack)


def handle_exceptions(func: Optional[Callable] = None, has_ui: bool = True) -> Callable:
    """
    Decorator for handling exceptions, which tries to find a matching pattern for known errors.
    If the pattern is found, the exception is handled according to the specified handler.
    Otherwise, the exception is raised as usual.

    :param func: Function to be decorated.
    :type func: Callable
    :returns: Decorated function.
    :rtype: Callable

    :Usage Example:

        .. code-block:: python

            import supervisely as sly

            @sly.handle_exceptions
            def my_func():
                # Some code that may raise an exception.
                raise Exception("Test exception")

            # call with argument `has_ui=False` if you don't want to raise a DialogWindowError
            @sly.handle_exceptions(has_ui=False)
            def my_func():
                # Some code that may raise an exception.
                raise Exception("Test exception")
    """

    f = None
    if func is None or callable(func):
        # No argument provided or used as a decorator without arguments
        f = func

    def decorator(f) -> Callable:
        @wraps(f)
        def wrapper(*args, **kwargs):
            try:
                return f(*args, **kwargs)
            except Exception as e:
                exception_handler = handle_exception(e)
                if exception_handler:
                    exception_handler.raise_error(has_ui=has_ui)
                else:
                    raise

        return wrapper

    return decorator if f is None else decorator(f)


def handle_additional_exceptions(
    errors: Dict[BaseException, Dict[str, HandleException]] = {},
) -> Callable:
    ERROR_PATTERNS.update(errors)

    def inner(func: Callable) -> Callable:
        def wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except Exception as e:
                exception_handler = handle_exception(e)
                if exception_handler:
                    exception_handler.raise_error()
                else:
                    raise

        return wrapper

    return inner


def read_stack_from_exception(exception):
    stack = traceback.extract_stack()
    try:
        stack.extend(traceback.extract_tb(exception.__traceback__))
    except AttributeError:
        pass
    return stack


def get_message_from_exception(exception, title):
    try:
        json_text = exception.args[0].response.text
        info = loads(json_text)
        exc_message = info.get("message", str(exception))
    except:
        exc_message = str(exception)

    return f"{title}: {exc_message}"



from supervisely.app import DataJson, StateJson
from supervisely.app.widgets import Widget
from typing import Union, List


class Slider(Widget):
    """Slider widget for selecting a numeric value or a range, with value-changed callback."""

    class Routes:
        """Callback route names used by the widget frontend to notify Python."""

        VALUE_CHANGED = "value_changed"

    def __init__(
        self,
        value: Union[int, List[int]] = 0,
        min: int = 0,
        max: int = 100,
        step: int = 1,
        show_input: bool = False,
        show_input_controls: bool = False,
        show_stops: bool = False,
        show_tooltip: bool = True,
        range: bool = False,  # requires value to be List[int, int]
        vertical: bool = False,
        height: int = None,
        widget_id: str = None,
        style: str = "padding: 20px",
    ):
        """:param value: Initial value (int or [min, max] if range=True).
        :type value: Union[int, List[int]]
        :param min: Minimum value.
        :type min: int
        :param max: Maximum value.
        :type max: int
        :param step: Step size.
        :type step: int
        :param show_input: If True, show numeric input.
        :type show_input: bool
        :param show_input_controls: If True, show +/- buttons.
        :type show_input_controls: bool
        :param show_stops: If True, show stop marks.
        :type show_stops: bool
        :param show_tooltip: If True, show value tooltip.
        :type show_tooltip: bool
        :param range: If True, select range [min, max].
        :type range: bool
        :param vertical: If True, vertical orientation.
        :type vertical: bool
        :param height: Height in px (for vertical).
        :type height: int, optional
        :param widget_id: Unique widget identifier.
        :type widget_id: str, optional
        :param style: CSS style string.
        :type style: str
        """
        self._value = value
        self._min = min
        self._max = max
        self._step = step
        self._show_input = show_input
        self._show_input_controls = show_input_controls if show_input else False
        self._show_stops = show_stops
        self._show_tooltip = show_tooltip
        self._range = False if show_input else range
        self._vertical = vertical
        self._height = f"{height}px" if vertical else None
        self._style = style
        self._changes_handled = False

        self._validate_min_max(min, max)
        self._validate_default_value(value)

        super().__init__(widget_id=widget_id, file_path=__file__)

    def _validate_min_max(self, min_val: int, max_val: int):
        if min_val > max_val:
            raise ValueError(
                f"Minimum value: '{min_val}' can't be bigger than maximum value: '{max_val}'"
            )

    def _validate_default_value(self, value: Union[int, float, List[int]]):
        if isinstance(value, int) or isinstance(value, float):
            if self._range:
                raise ValueError(f"value = '{value}', should be 'list' if range is True")
            if value < self._min:
                self._value = self._min
            elif value > self._max:
                self._value = self._max
        elif isinstance(value, list):
            if not self._range:
                raise ValueError(f"value = '{value}', should be 'int' if range is False")
            if value[0] < self._min and value[1] > self._max:
                self._value = self._min
                self._value = self._max
            elif value[0] < self._min and value[1] < self._min:
                self._value[0] = self._min
                self._value[1] = self._min
            elif value[0] > self._max and value[1] > self._max:
                self._value[0] = self._max
                self._value[1] = self._max
        else:
            raise ValueError(
                f"value = '{value}', should be 'int', 'float' or 'list', not '{type(value)}'"
            )

    def get_json_data(self):
        return {
            "min": self._min,
            "max": self._max,
            "step": self._step,
            "showInput": self._show_input,
            "showInputControls": self._show_input_controls,
            "showStops": self._show_stops,
            "showTooltip": self._show_tooltip,
            "range": self._range,
            "vertical": self._vertical,
            "height": self._height,
        }

    def get_json_state(self):
        return {"value": self._value}

    def set_value(self, value: Union[int, List[int]]):
        self._validate_default_value(value)
        self._value = value
        StateJson()[self.widget_id]["value"] = self._value
        StateJson().send_changes()

    def get_value(self):
        return StateJson()[self.widget_id]["value"]

    def set_min(self, value: int):
        self._validate_min_max(value, self.get_max())
        DataJson()[self.widget_id]["min"] = value
        DataJson().send_changes()

    def get_min(self):
        return DataJson()[self.widget_id]["min"]

    def set_max(self, value: int):
        self._validate_min_max(self.get_min(), value)
        DataJson()[self.widget_id]["max"] = value
        DataJson().send_changes()

    def get_max(self):
        return DataJson()[self.widget_id]["max"]

    def set_step(self, value: int):
        DataJson()[self.widget_id]["step"] = value
        DataJson().send_changes()

    def get_step(self):
        return DataJson()[self.widget_id]["step"]

    def is_input_enabled(self):
        return DataJson()[self.widget_id]["showInput"]

    def show_input(self):
        DataJson()[self.widget_id]["showInput"] = True
        DataJson().send_changes()

    def hide_input(self):
        DataJson()[self.widget_id]["showInput"] = False
        DataJson().send_changes()

    def is_input_controls_enabled(self):
        return DataJson()[self.widget_id]["showInputControls"]

    def show_input_controls(self):
        DataJson()[self.widget_id]["showInputControls"] = True
        DataJson().send_changes()

    def hide_input_controls(self):
        DataJson()[self.widget_id]["showInputControls"] = False
        DataJson().send_changes()

    def is_step_enabled(self):
        return DataJson()[self.widget_id]["showStops"]

    def show_steps(self):
        DataJson()[self.widget_id]["showStops"] = True
        DataJson().send_changes()

    def hide_steps(self):
        DataJson()[self.widget_id]["showStops"] = False
        DataJson().send_changes()

    def is_tooltip_enabled(self):
        return DataJson()[self.widget_id]["showTooltip"]

    def show_tooltip(self):
        DataJson()[self.widget_id]["showTooltip"] = True
        DataJson().send_changes()

    def hide_tooltip(self):
        DataJson()[self.widget_id]["showTooltip"] = False
        DataJson().send_changes()

    def value_changed(self, func):
        route_path = self.get_route_path(Slider.Routes.VALUE_CHANGED)
        server = self._sly_app.get_server()
        self._changes_handled = True

        @server.post(route_path)
        def _click():
            res = self.get_value()
            self._value = res
            func(res)

        return _click



import numpy as np
from typing import Callable


class LossPlateauDetector:
    """Detect plateau in training loss using moving average comparison."""

    def __init__(
        self,
        window_size: int = 150,
        threshold: float = 0.002,
        patience: int = 3,
        check_interval: int = 1,
    ):
        """:param window_size: Number of iterations for moving average.
        :type window_size: int
        :param threshold: Relative change threshold (e.g., 0.005 = 0.5%).
        :type threshold: float
        :param patience: Number of consecutive plateau detections before action.
        :type patience: int
        :param check_interval: Check frequency (every N iterations).
        :type check_interval: int
        """
        self.window_size = window_size
        self.threshold = threshold
        self.check_interval = check_interval
        self.patience = patience
        self._min_iterations = 2 * window_size

        # State
        self.loss_history = []
        self.consecutive_plateau_count = 0
        self._save_checkpoint_fn = None

    def register_save_checkpoint_callback(self, fn: Callable[[], None]):
        """Register callback function to save checkpoint when plateau detected"""
        self._save_checkpoint_fn = fn

    def reset(self):
        """Reset detector state"""
        self.loss_history = []
        self.consecutive_plateau_count = 0

    def step(self, loss: float, current_iter: int) -> bool:
        """
        Process one training iteration.

        :param loss: Current loss value
        :type loss: float
        :param current_iter: Current iteration number
        :type current_iter: int
        :returns: True if plateau confirmed and checkpoint saved, False otherwise
        :rtype: bool
        """
        self.loss_history.append(loss)

        # Check only at specified intervals
        if (current_iter + 1) % self.check_interval != 0:
            return False

        # Need enough data
        if len(self.loss_history) < self._min_iterations:
            return False

        # Check for plateau
        is_plateau, info = self._check_plateau(current_iter)

        if is_plateau:
            self.consecutive_plateau_count += 1
            print(
                f'[Plateau Detection] Iteration {current_iter}: '
                f'Signal {self.consecutive_plateau_count}/{self.patience} '
                f'(change: {info["metric"]:.6f}, threshold: {self.threshold})'
            )

            # Trigger action when patience reached
            if self.consecutive_plateau_count >= self.patience:
                print(f'[Plateau Detection] Plateau confirmed, saving checkpoint...')

                if self._save_checkpoint_fn is not None:
                    self._save_checkpoint_fn()
                    print(f'[Plateau Detection] Checkpoint saved')
                else:
                    print(f'[Plateau Detection] No callback registered')

                self.consecutive_plateau_count = 0
                return True

        return False

    def _check_plateau(self, current_iter: int) -> tuple:
        """Check if current window shows plateau"""
        # Current window average
        current_window = self.loss_history[-self.window_size:]
        current_avg = np.mean(current_window)

        # Previous window average
        previous_window = self.loss_history[-2*self.window_size:-self.window_size]
        previous_avg = np.mean(previous_window)

        change = previous_avg - current_avg
        is_plateau = change < self.threshold

        info = {
            'iter': current_iter,
            'metric': change,
            'threshold': self.threshold,
            'previous_avg': previous_avg,
            'current_avg': current_avg,
        }

        return is_plateau, info



from supervisely.nn.artifacts.artifacts import BaseTrainArtifacts
from supervisely.nn.artifacts import (
    YOLOv5,
    YOLOv5v2,
    YOLOv8,
    MMClassification,
    MMPretrain,
    MMSegmentation,
    MMDetection,
    MMDetection3,
    Detectron2,
    UNet,
    RITM,
    RTDETR,
)


class FrameworkName:
    """String constants for supported training frameworks."""

    YOLOV5 = "YOLOv5"
    YOLOV5V2 = "YOLOv5 2.0"
    YOLOV8 = "YOLOv8+"
    MMCLASSIFICATION = "MMClassification"
    MMPRETRAIN = "MMPretrain"
    MMSEGMENTATION = "MMSegmentation"
    MMDETECTION = "MMDetection"
    MMDETECTION3 = "MMDetection 3.0"
    RTDETR = "RT-DETR"
    DETECTRON2 = "Detectron2"
    UNET = "UNet"
    RITM = "RITM"


class FrameworkMapper:
    """Factory that maps a framework name to its :class:`~supervisely.nn.artifacts.artifacts.BaseTrainArtifacts` class."""

    _map = {
        FrameworkName.YOLOV5: YOLOv5,
        FrameworkName.YOLOV5V2: YOLOv5v2,
        FrameworkName.YOLOV8: YOLOv8,
        FrameworkName.MMCLASSIFICATION: MMClassification,
        FrameworkName.MMPRETRAIN: MMPretrain,
        FrameworkName.MMSEGMENTATION: MMSegmentation,
        FrameworkName.MMDETECTION: MMDetection,
        FrameworkName.MMDETECTION3: MMDetection3,
        FrameworkName.RTDETR: RTDETR,
        FrameworkName.DETECTRON2: Detectron2,
        FrameworkName.UNET: UNet,
        FrameworkName.RITM: RITM,
    }

    @classmethod
    def get_framework_cls(cls, name: str, team_id: int) -> BaseTrainArtifacts:
        if name not in cls._map:
            raise ValueError(f"Unknown framework: {name}")
        return cls._map[name](team_id)



from __future__ import annotations

import os
import pickle
from pathlib import Path

import pandas as pd

from supervisely.io.json import dump_json_file, load_json_file
from supervisely.nn.benchmark.base_evaluator import BaseEvalResult, BaseEvaluator
from supervisely.nn.benchmark.object_detection.metric_provider import MetricProvider
from supervisely.nn.benchmark.utils import (
    calculate_metrics,
    read_coco_datasets,
    sly2coco,
)
from supervisely.nn.benchmark.visualization.vis_click_data import ClickData, IdMapper
from supervisely.sly_logger import logger


class ObjectDetectionEvalResult(BaseEvalResult):
    """Evaluation result for object detection benchmarks (COCO bbox metrics + click-data helpers)."""

    mp_cls = MetricProvider
    PRIMARY_METRIC = "mAP"

    def _read_files(self, path: str) -> None:
        """Read all necessary files from the directory"""

        self.coco_gt = Path(path) / "cocoGt.json"
        self.coco_dt = Path(path) / "cocoDt.json"

        if self.coco_gt.exists() and self.coco_dt.exists():
            self.coco_gt, self.coco_dt = read_coco_datasets(self.coco_gt, self.coco_dt)

        eval_data_pickle_path = Path(path) / "eval_data.pkl"
        eval_data_archive_path = Path(path) / "eval_data.zip"
        if eval_data_pickle_path.exists():
            try:
                with open(Path(path, "eval_data.pkl"), "rb") as f:
                    self.eval_data = pickle.load(f)
            except Exception as e:
                logger.warning(f"Failed to load eval_data.pkl: {e}", exc_info=True)
                self.eval_data = None
        if self.eval_data is None and eval_data_archive_path.exists():
            try:
                self.eval_data = self._load_eval_data_archive(eval_data_archive_path)
            except Exception as e:
                logger.warning(f"Failed to load eval_data from archive: {e}", exc_info=True)
                self.eval_data = None

        if self.eval_data is None:
            raise ValueError("Failed to load eval_data. Please contact support.")

        inference_info_path = Path(path) / "inference_info.json"
        if inference_info_path.exists():
            self.inference_info = load_json_file(str(inference_info_path))

        speedtest_info_path = Path(path).parent / "speedtest" / "speedtest.json"
        if speedtest_info_path.exists():
            self.speedtest_info = load_json_file(str(speedtest_info_path))

    def _prepare_data(self) -> None:
        """Prepare data to allow easy access to the most important parts"""

        from pycocotools.coco import COCO  # pylint: disable=import-error

        if not hasattr(self, "coco_gt") or not hasattr(self, "coco_dt"):
            raise ValueError("GT and DT datasets are not provided")

        if not isinstance(self.coco_gt, COCO) and not isinstance(self.coco_dt, COCO):
            self.coco_gt, self.coco_dt = read_coco_datasets(self.coco_gt, self.coco_dt)

        self.mp = MetricProvider(
            self.eval_data,
            self.coco_gt,
            self.coco_dt,
        )
        self.mp.calculate()

        self.df_score_profile = pd.DataFrame(
            self.mp.confidence_score_profile(), columns=["scores", "precision", "recall", "f1"]
        )

        # downsample
        if len(self.df_score_profile) > 5000:
            self.dfsp_down = self.df_score_profile.iloc[:: len(self.df_score_profile) // 1000]
        else:
            self.dfsp_down = self.df_score_profile

        # Click data
        gt_id_mapper = IdMapper(self.coco_gt.dataset)
        dt_id_mapper = IdMapper(self.coco_dt.dataset)

        self.click_data = ClickData(self.mp.m, gt_id_mapper, dt_id_mapper)

    @classmethod
    def from_evaluator(cls, evaulator: ObjectDetectionEvaluator) -> ObjectDetectionEvalResult:
        """Method to customize loading of the evaluation result."""
        eval_result = cls()
        eval_result.eval_data = evaulator.eval_data
        eval_result.coco_gt = evaulator.cocoGt
        eval_result.coco_dt = evaulator.cocoDt
        eval_result._prepare_data()
        return eval_result

    @property
    def key_metrics(self):
        return self.mp.key_metrics()

    @property
    def different_iou_thresholds_per_class(self) -> bool:
        return self.mp.iou_threshold_per_class is not None


class ObjectDetectionEvaluator(BaseEvaluator):
    """Benchmark evaluator that computes object-detection metrics between GT and prediction projects."""

    EVALUATION_PARAMS_YAML_PATH = f"{Path(__file__).parent}/evaluation_params.yaml"
    eval_result_cls = ObjectDetectionEvalResult
    accepted_shapes = ["rectangle"]

    def evaluate(self):
        try:
            self.cocoGt_json, self.cocoDt_json = self._convert_to_coco()
        except AssertionError as e:
            raise ValueError(
                f"{e}. Please make sure that your GT and DT projects are correct. "
                "If GT project has nested datasets and DT project was created with NN app, "
                "try to use newer version of NN app."
            )
        self.cocoGt, self.cocoDt = read_coco_datasets(self.cocoGt_json, self.cocoDt_json)
        with self.pbar(message="Evaluation: Calculating metrics", total=5) as p:
            self.eval_data = calculate_metrics(
                self.cocoGt,
                self.cocoDt,
                iouType="bbox",
                progress_cb=p.update,
                evaluation_params=self.evaluation_params,
            )
        self._dump_eval_results()

    @classmethod
    def validate_evaluation_params(cls, evaluation_params: dict) -> None:
        available_thres = [0.5, 0.55, 0.6, 0.65, 0.7, 0.75, 0.8, 0.85, 0.9, 0.95]
        iou_threshold = evaluation_params.get("iou_threshold")
        if iou_threshold is not None:
            assert iou_threshold in available_thres, (
                f"iou_threshold must be one of {available_thres}, " f"but got {iou_threshold}"
            )
        iou_threshold_per_class = evaluation_params.get("iou_threshold_per_class")
        if iou_threshold_per_class is not None:
            for class_name, iou_thres in iou_threshold_per_class.items():
                assert iou_thres in available_thres, (
                    f"class {class_name}: iou_threshold_per_class must be one of {available_thres}, "
                    f"but got {iou_thres}"
                )

    def _convert_to_coco(self):
        cocoGt_json = sly2coco(
            self.gt_project_path,
            is_dt_dataset=False,
            accepted_shapes=self.accepted_shapes,
            progress=self.pbar,
            classes_whitelist=self.classes_whitelist,
        )
        cocoDt_json = sly2coco(
            self.pred_project_path,
            is_dt_dataset=True,
            accepted_shapes=self.accepted_shapes,
            progress=self.pbar,
            classes_whitelist=self.classes_whitelist,
        )

        if len(cocoGt_json["annotations"]) == 0:
            raise ValueError("Not found any annotations in GT project")
        if len(cocoDt_json["annotations"]) == 0:
            raise ValueError(
                "Not found any predictions. "
                "Please make sure that your model produces predictions."
            )
        assert (
            cocoDt_json["categories"] == cocoGt_json["categories"]
        ), "Classes in GT and Pred projects must be the same"
        assert [f'{x["dataset"]}/{x["file_name"]}' for x in cocoDt_json["images"]] == [
            f'{x["dataset"]}/{x["file_name"]}' for x in cocoGt_json["images"]
        ], "Images in GT and DT projects are different"
        return cocoGt_json, cocoDt_json

    def _dump_eval_results(self):
        cocoGt_path, cocoDt_path, eval_data_path = self._get_eval_paths()
        dump_json_file(self.cocoGt_json, cocoGt_path, indent=None)
        dump_json_file(self.cocoDt_json, cocoDt_path, indent=None)
        self._dump_eval_results_archive()
        self._dump_pickle(self.eval_data, eval_data_path)

    def _get_eval_paths(self):
        base_dir = self.result_dir
        cocoGt_path = os.path.join(base_dir, "cocoGt.json")
        cocoDt_path = os.path.join(base_dir, "cocoDt.json")
        eval_data_path = os.path.join(base_dir, "eval_data.pkl")
        return cocoGt_path, cocoDt_path, eval_data_path



from typing import Dict, Union
from supervisely.app import DataJson
from supervisely.app.widgets import Widget
from supervisely import is_development


class TaskLogs(Widget):
    """Widget that displays live logs for a Supervisely task (job) by task ID."""

    def __init__(
        self,
        task_id: int = None,
        widget_id: str = None,
    ):
        """:param task_id: Supervisely task (job) ID to display logs for.
        :type task_id: int, optional
        :param widget_id: Unique widget identifier.
        :type widget_id: str, optional
        """
        self._task_id = task_id
        self._is_development = is_development()
        super().__init__(widget_id=widget_id, file_path=__file__)

    def get_json_data(self) -> Dict:
        return {"taskId": self._task_id}

    def get_json_state(self) -> Dict:
        return {}

    def get_task_id(self) -> int:
        return DataJson()[self.widget_id]["taskId"]

    def _set_task_id(self, task_id: Union[int, None]):
        self._task_id = task_id
        DataJson()[self.widget_id]["taskId"] = self._task_id
        DataJson().send_changes()

    def set_task_id(self, task_id: int):
        self._set_task_id(None)
        if type(task_id) != int:
            raise TypeError(f"task_id must be int, but {type(task_id)} was given")
        self._set_task_id(task_id)



from __future__ import annotations

from typing import Dict, List, Optional

from supervisely.app import DataJson
from supervisely.app.widgets import Widget

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal


class ActivityFeed(Widget):
    """Vertical list of activity items with status indicators (pending, in progress, completed, failed)."""

    class Item:
        """Single item in ActivityFeed (content, status, optional number)."""

        def __init__(
            self,
            content: Widget,
            status: Literal["pending", "in_progress", "completed", "failed"] = "pending",
            number: Optional[int] = None,
        ) -> ActivityFeed.Item:
            """
            :param content: Widget to display as item content.
            :type content: :class:`~supervisely.app.widgets.widget.Widget`
            :param status: pending, in_progress, completed, or failed.
            :type status: Literal["pending", "in_progress", "completed", "failed"]
            :param number: Position in feed (auto-assigned if None).
            :type number: Optional[int]
            """
            self.content = content
            self.status = status
            self.number = number
            self._validate_status()

        def _validate_status(self):
            valid_statuses = ["pending", "in_progress", "completed", "failed"]
            if self.status not in valid_statuses:
                raise ValueError(
                    f"Invalid status '{self.status}'. Must be one of: {', '.join(valid_statuses)}"
                )

        def to_json(self):
            return {
                "number": self.number,
                "status": self.status,
            }

    def __init__(
        self,
        items: Optional[List[ActivityFeed.Item]] = None,
        widget_id: Optional[str] = None,
    ):
        """
        :param items: List of ActivityFeed.Item to display. Items auto-numbered if number not set.
        :type items: Optional[List[:class:`~supervisely.app.widgets.activity_feed.activity_feed.ActivityFeed.Item`]]
        :param widget_id: Widget identifier.
        :type widget_id: str, optional

        :Usage Example:

            .. code-block:: python

                from supervisely.app.widgets import ActivityFeed, Text

                item1 = ActivityFeed.Item(content=Text("Processing"), status="completed")
                item2 = ActivityFeed.Item(content=Text("Training"), status="in_progress", number=2)
                feed = ActivityFeed(items=[item1, item2])
                feed.set_status(2, "completed")
        """
        self._items = items if items is not None else []
        self._auto_assign_numbers()
        super().__init__(widget_id=widget_id, file_path=__file__)

    def _auto_assign_numbers(self):
        """Automatically assign numbers to items that don't have them."""
        next_number = 1
        for item in self._items:
            if item.number is None:
                item.number = next_number
            next_number = max(next_number, item.number) + 1

    def get_json_data(self) -> Dict:
        """Returns dictionary with widget data.

        :returns: Dictionary with items data
        :rtype: Dict
        """
        return {
            "items": [item.to_json() for item in self._items],
        }

    def get_json_state(self) -> Dict:
        """Returns dictionary with widget state (empty for this widget).

        :returns: Empty dictionary
        :rtype: Dict
        """
        return {}

    def add_item(
        self,
        item: Optional[ActivityFeed.Item] = None,
        content: Optional[Widget] = None,
        status: Literal["pending", "in_progress", "completed", "failed"] = "pending",
        number: Optional[int] = None,
    ) -> None:
        """Add a new item to the activity feed.

        You can either pass an item object or provide content and status separately.

        :param item: Item to add
        :type item: Optional[:class:`~supervisely.app.widgets.activity_feed.activity_feed.ActivityFeed.Item`]
        :param content: Widget content (used if item is not provided)
        :type content: Optional[:class:`~supervisely.app.widgets.widget.Widget`]
        :param status: Status of the item (used if item is not provided)
        :type status: Literal["pending", "in_progress", "completed", "failed"]
        :param number: Position number (auto-assigned if not provided)
        :type number: Optional[int]
        """
        if item is None:
            if content is None:
                raise ValueError("Either 'item' or 'content' must be provided")
            item = ActivityFeed.Item(content=content, status=status, number=number)

        if item.number is None:
            # Auto-assign number
            if self._items:
                item.number = max(i.number for i in self._items) + 1
            else:
                item.number = 1

        self._items.append(item)
        self.update_data()
        DataJson().send_changes()

    def remove_item(self, number: int) -> None:
        """Remove an item from the activity feed by its number.

        :param number: Number of the item to remove. Starts from 1.
        :type number: int
        """
        self._items = [item for item in self._items if item.number != number]
        self.update_data()
        DataJson().send_changes()

    def set_status(
        self,
        number: int,
        status: Literal["pending", "in_progress", "completed", "failed"],
    ) -> None:
        """Update the status of an item by its number.

        :param number: Number of the item to update. Starts from 1.
        :type number: int
        :param status: New status for the item
        :type status: Literal["pending", "in_progress", "completed", "failed"]
        """
        for i, item in enumerate(self._items):
            if item.number == number:
                item.status = status
                item._validate_status()
                DataJson()[self.widget_id]["items"][i]["status"] = status
                DataJson().send_changes()
                return
        raise ValueError(f"Item with number {number} not found")

    def get_status(self, number: int) -> str:
        """Get the status of an item by its number.

        :param number: Number of the item. Starts from 1.
        :type number: int
        :returns: Status of the item
        :rtype: str
        """
        for item in self._items:
            if item.number == number:
                return item.status
        raise ValueError(f"Item with number {number} not found")

    def get_items(self) -> List[ActivityFeed.Item]:
        """Get all items in the activity feed.

        :returns: List of all items
        :rtype: List[:class:`~supervisely.app.widgets.activity_feed.activity_feed.ActivityFeed.Item`]
        """
        return self._items

    def clear(self) -> None:
        """Remove all items from the activity feed."""
        self._items = []
        self.update_data()
        DataJson().send_changes()

    def set_items(self, items: List[ActivityFeed.Item]) -> None:
        """Replace all items in the activity feed.

        :param items: New list of items
        :type items: List[:class:`~supervisely.app.widgets.activity_feed.activity_feed.ActivityFeed.Item`]
        """
        self._items = items
        self._auto_assign_numbers()
        self.update_data()
        DataJson().send_changes()



import os
import uuid
from pathlib import Path
from urllib3.util import parse_url
from typing import List, Optional, Tuple

import supervisely as sly
from supervisely.app import StateJson, DataJson
from supervisely.app.widgets import Button, FolderThumbnail, GridGallery, Slider, Widget
from supervisely.app.fastapi.offline import get_offline_session_files_path


class ImagePairSequence(Widget):
    """Gallery of left/right image pairs from Team Files with slider to browse pairs and sync view."""

    def __init__(
        self,
        opacity: Optional[float] = 0.4,
        enable_zoom: Optional[bool] = False,
        sync_views: Optional[bool] = True,
        slider_title: Optional[str] = "pairs",
        widget_id=None,
    ):
        """
        :param opacity: Opacity of annotation overlays (0-1).
        :type opacity: float, optional
        :param enable_zoom: If True, enable zoom.
        :type enable_zoom: bool, optional
        :param sync_views: If True, sync views.
        :type sync_views: bool, optional
        :param slider_title: Title of the slider.
        :type slider_title: str, optional
        :param widget_id: Unique widget identifier.
        :type widget_id: str, optional
        """
        self._api = sly.Api.from_env()
        self._team_id = sly.env.team_id()

        # init data for gallery
        self._left_data = []
        self._right_data = []

        # init gallery options
        self._columns = 2
        self._current_grid = 0
        self._total_grids = 0
        self._slider_title = slider_title
        self._need_update = False

        self._first_button = Button(
            "", icon="zmdi zmdi-skip-previous", plain=True, button_size="mini"
        )
        self._prev_button = Button(
            "", icon="zmdi zmdi-chevron-left", plain=True, button_size="mini"
        )
        self._next_button = Button(
            "", icon="zmdi zmdi-chevron-right", plain=True, button_size="mini"
        )
        self._last_button = Button("", icon="zmdi zmdi-skip-next", plain=True, button_size="mini")

        self._slider = Slider(show_stops=True, min=1, max=1, step=1, value=1)

        self._folder_thumbnail = FolderThumbnail()

        self._grid_gallery = GridGallery(
            columns_number=self._columns,
            annotations_opacity=opacity,
            show_opacity_slider=True,
            enable_zoom=enable_zoom,
            resize_on_zoom=False,
            sync_views=sync_views,
            fill_rectangle=False,
        )
        self._grid_gallery.hide()

        @self._slider.value_changed
        def on_slider_change(value):
            self._update_gallery(int(value))

        @self._first_button.click
        def on_first_button_click():
            if self._current_grid > 1:
                self._update_gallery(1)

        @self._prev_button.click
        def on_prev_button_click():
            if self._current_grid > 1:
                self._update_gallery(self._current_grid - 1)

        @self._next_button.click
        def on_next_button_click():
            if self._current_grid < self._total_grids:
                self._update_gallery(self._current_grid + 1)

        @self._last_button.click
        def on_last_button_click():
            if self._current_grid < self._total_grids:
                self._update_gallery(self._total_grids)

        super().__init__(widget_id=widget_id, file_path=__file__)

    def get_json_data(self) -> dict:
        return {
            "leftAnnotations": [],
            "rightAnnotations": [],
        }

    def get_json_state(self) -> dict:
        return {
            "currentGrid": self._current_grid,
            "totalGrids": self._total_grids,
            "sliderTitle": f" {self._slider_title}",
        }

    def append_left(self, path: str, ann: sly.Annotation = None, title: str = None):
        self._add_with_check("left", [[path, ann, title]])
        self._update_data()

    def append_right(self, path: str, ann: sly.Annotation = None, title: str = None):
        self._add_with_check("right", [[path, ann, title]])
        self._update_data()

    def extend_left(
        self, paths: List[str], anns: List[sly.Annotation] = None, titles: List[str] = None
    ):
        anns = [None] * len(paths) if anns is None else anns
        titles = [None] * len(paths) if titles is None else titles
        data = list(zip(paths, anns, titles))
        self._add_with_check("left", data)
        self._update_data()

    def extend_right(
        self, paths: List[str], anns: List[sly.Annotation] = None, titles: List[str] = None
    ):
        anns = [None] * len(paths) if anns is None else anns
        titles = [None] * len(paths) if titles is None else titles
        data = list(zip(paths, anns, titles))
        self._add_with_check("right", data)
        self._update_data()

    def append_pair(
        self,
        left: Tuple[str, Optional[sly.Annotation], Optional[str]],
        right: Tuple[str, Optional[sly.Annotation], Optional[str]],
    ):
        self._add_with_check("left", [left])
        self._add_with_check("right", [right])
        self._update_data()

    def extend_pairs(self, left: List[Tuple], right: List[Tuple]):
        self._add_with_check("left", left)
        self._add_with_check("right", right)
        self._update_data()

    def clean_up(self):
        self._left_data = []
        self._right_data = []
        self._grid_gallery.hide()
        self._slider.set_value(1)
        self._slider.set_max(1)
        self._need_update = False
        self._current_grid = 0
        self._total_grids = 0
        StateJson()[self.widget_id]["totalGrids"] = 0
        StateJson()[self.widget_id]["currentGrid"] = 0
        StateJson().send_changes()

    def _check_paths(self, paths):
        for path in paths:
            parsed_path = parse_url(path)
            if parsed_path.scheme not in (None, "http", "https"):
                raise ValueError(f"Invalid path or url to image: {path}")

    def _prepare_annotations(self, anns):
        new_anns = []
        for ann in anns:
            if ann is not None:
                new_labels = []
                for label in ann.labels:
                    if label.geometry.sly_id is None:
                        label.geometry.sly_id = str(uuid.uuid4().int)
                    new_labels.append(label)
                ann = ann.clone(labels=new_labels)
            new_anns.append(ann)
        return new_anns

    def _add_with_check(self, side, data):
        paths, anns, titles = [], [], []
        for item in data:
            if len(item) < 3:
                item = item + (None,) * (3 - len(item))
            path, ann, title = item
            paths.append(path)
            anns.append(ann)
            titles.append(title)

        self._check_paths(paths)
        new_anns = self._prepare_annotations(anns)

        total_grids = max(len(self._left_data), len(self._right_data), 1)
        if self._total_grids != total_grids:
            self._need_update = True

        has_empty_before = any([len(self._left_data) == 0, len(self._right_data) == 0])

        ann_jsons = [a.to_json() if a is not None else None for a in new_anns]
        data = list(zip(paths, new_anns, titles))
        if side == "left":
            self._left_data.extend(data)
            DataJson()[self.widget_id]["leftAnnotations"].extend(ann_jsons)
        elif side == "right":
            self._right_data.extend(data)
            DataJson()[self.widget_id]["rightAnnotations"].extend(ann_jsons)
        DataJson().send_changes()

        has_empty_after = any([len(self._left_data) == 0, len(self._right_data) == 0])
        if has_empty_before and not has_empty_after:
            self._need_update = True

        self._dump_image_to_offline_sessions_file(paths)

    def _dump_image_to_offline_sessions_file(self, paths: List[str]):
        if sly.is_production():
            remote_dir = get_offline_session_files_path(self._api.task_id)
            remote_dir = remote_dir.joinpath("sly", "css", "app", "widgets", "image_pair_sequence")
            dst_paths = [remote_dir.joinpath(Path(path).name).as_posix() for path in paths]
            local_paths = [self._download_image(path) for path in paths]

            self._api.file.upload_bulk(
                team_id=self._team_id,
                src_paths=local_paths,
                dst_paths=dst_paths,
            )

    def _download_image(self, path: str):
        if path.lstrip("/").startswith("static/"):
            path = path.lstrip("/")[len("static/") :]
            app = sly.Application()
            save_path = os.path.join(app.get_static_dir(), path)
            if not os.path.exists(save_path):
                raise FileNotFoundError(f"File {save_path} not found")
        else:
            save_path = os.path.join(sly.app.get_data_dir(), sly.fs.get_file_name_with_ext(path))
            sly.fs.download(path, save_path)
        return save_path

    def _update_data(self):
        self._total_grids = max(len(self._left_data), len(self._right_data), 1)
        min_len = min(len(self._left_data), len(self._right_data))
        if self._slider.get_value() == self._slider.get_max():
            self._need_update = True

        self._slider.set_max(self._total_grids)

        if self._need_update and min_len > 0:
            self._update_gallery(self._total_grids)

        StateJson()[self.widget_id]["totalGrids"] = self._total_grids
        StateJson().send_changes()

    def _update_gallery(self, page: int):
        if self._grid_gallery.is_hidden():
            self._grid_gallery.show()
        self._slider.set_value(page)
        self._current_grid = page
        StateJson()[self.widget_id]["currentGrid"] = page
        StateJson().send_changes()

        self._grid_gallery.clean_up()

        len_left = len(self._left_data)
        len_right = len(self._right_data)

        if len_left > 0 and len_right > 0:
            left = self._left_data[page - 1] if page <= len_left else self._left_data[-1]
            self._grid_gallery.append(*left)  # set left

            right = self._right_data[page - 1] if page <= len_right else self._right_data[-1]
            self._grid_gallery.append(*right)  # set right

        DataJson().send_changes()
        self._need_update = False

    def disable(self):
        self._disabled = True
        self._grid_gallery.disable()
        self._first_button.disable()
        self._prev_button.disable()
        self._next_button.disable()
        self._last_button.disable()
        self._slider.disable()
        DataJson()[self.widget_id]["disabled"] = self._disabled
        DataJson().send_changes()

    def enable(self):
        self._disabled = False
        self._grid_gallery.enable()
        self._first_button.enable()
        self._prev_button.enable()
        self._next_button.enable()
        self._last_button.enable()
        self._slider.enable()
        DataJson()[self.widget_id]["disabled"] = self._disabled
        DataJson().send_changes()



from typing import Dict, List

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

from supervisely.app import StateJson, DataJson
from supervisely.app.widgets import Widget
from supervisely.api.api import Api


class TeamFilesSelector(Widget):
    """File/folder picker for Team Files; supports single/multiple selection and optional file type filter."""

    def __init__(
        self,
        team_id: int,
        multiple_selection: bool = False,
        max_height: int = 500,
        selection_file_type: Literal["folder", "file"] = None,
        hide_header: bool = True,
        hide_empty_table: bool = True,
        additional_fields: List[
            Literal["id", "createdAt", "updatedAt", "type", "size", "mimeType"]
        ] = [],
        widget_id: str = None,
        initial_folder: str = None,
    ):
        """:param team_id: Team ID for Team Files.
        :type team_id: int
        :param multiple_selection: If True, allow multiple selection.
        :type multiple_selection: bool
        :param max_height: Max height in pixels.
        :type max_height: int
        :param selection_file_type: Filter: "folder" or "file".
        :type selection_file_type: Literal["folder", "file"], optional
        :param hide_header: If True, hide table header.
        :type hide_header: bool
        :param hide_empty_table: If True, hide when empty.
        :type hide_empty_table: bool
        :param additional_fields: Extra columns to show.
        :type additional_fields: List[Literal]
        :param widget_id: Unique widget identifier.
        :type widget_id: str, optional
        :param initial_folder: Initial folder path.
        :type initial_folder: str, optional

        :raises ValueError: If additional_fields contains invalid field names.
        """
        self._api = Api()
        self._team_id = team_id

        self._multiple_selection = multiple_selection
        self._max_height = f"{max_height}px"
        self._selection_file_type = selection_file_type
        self._hide_header = hide_header
        self._hide_empty_table = hide_empty_table

        available_fields = ["id", "createdAt", "updatedAt", "type", "size", "mimeType"]
        for field in additional_fields:
            if field not in available_fields:
                raise ValueError(
                    f'"{field}" is not a valid field. Available fields: {available_fields}.'
                )

        self._additional_fields = additional_fields
        self._selected = []
        self._initial_folder = initial_folder or "/"

        super().__init__(widget_id=widget_id, file_path=__file__)

    def get_json_data(self) -> Dict:
        return {
            "teamId": self._team_id,
            "options": {
                "multipleSelection": self._multiple_selection,
                "maxHeight": self._max_height,
                "selectionFileType": self._selection_file_type,
                "hideHeader": self._hide_header,
                "hideEmptyTable": self._hide_empty_table,
                "additionalFields": self._additional_fields,
                "initialFolder": {"path": self._initial_folder},
            },
        }

    def get_json_state(self) -> Dict:
        return {"selected": self._selected}

    def get_selected_items(self) -> List[dict]:
        selected = StateJson()[self.widget_id]["selected"]
        return [item for item in selected]

    def get_selected_paths(self) -> List[str]:
        selected = StateJson()[self.widget_id]["selected"]
        return [item["path"] for item in selected]

    def set_team_id(self, team_id: int):
        if type(team_id) != int:
            raise ValueError(f"team_id must be int, got {type(team_id)}")
        if team_id == self._team_id:
            return
        DataJson()[self.widget_id]["teamId"] = None
        DataJson().send_changes()
        self._team_id = team_id
        DataJson()[self.widget_id]["teamId"] = self._team_id
        DataJson().send_changes()



# coding: utf-8

from __future__ import annotations

from collections import Counter
from typing import Dict, Iterator, List, Optional

from supervisely.annotation.renamer import Renamer
from supervisely.annotation.tag_meta import TagMeta, TagValueType
from supervisely.collection.key_indexed_collection import KeyIndexedCollection
from supervisely.io.json import JsonSerializable


class TagMetaCollection(KeyIndexedCollection, JsonSerializable):
    """Collection with :class:`~supervisely.annotation.tag_meta.TagMeta` instances."""

    item_type = TagMeta

    def __init__(self, items: Optional[List[TagMeta]] = None):
        """
        :param items: List of TagMetas.
        :type items: List[:class:`~supervisely.annotation.tag_meta.TagMeta`]

        :raises :class:`~supervisely.collection.key_indexed_collection.DuplicateKeyError`: if instance with given name already exists

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                # Create TagMetas
                meta_weather = sly.TagMeta('Weather', sly.TagValueType.ANY_STRING)

                season_values = ["Winter", "Spring", "Summer", "Autumn"]
                meta_season = sly.TagMeta('Season', sly.TagValueType.ONEOF_STRING, possible_values=season_values)

                # Create TagMetaCollection from TagMetas
                tag_metas = sly.TagMetaCollection([meta_weather, meta_season])

                # Add items to TagMetaCollection
                meta_potato = sly.TagMeta('potato', sly.TagValueType.NONE)

                # Remember that TagMetaCollection is immutable, and we need to assign new instance of TagMetaCollection to a new variable
                tag_metas = tag_metas.add(meta_potato)

                # You can also add multiple items to collection
                meta_cabbage = sly.TagMeta('cabbage', sly.TagValueType.NONE)
                meta_carrot = sly.TagMeta('carrot', sly.TagValueType.NONE)
                meta_turnip = sly.TagMeta('turnip', sly.TagValueType.NONE)

                tag_metas = tag_metas.add_items([meta_cabbage, meta_carrot, meta_turnip])

                # Has key, checks if given key exist in collection
                tag_metas.has_key("cabbage")
                # Output: True

                # Intersection, finds intersection of given list of instances with collection items
                meta_dog = sly.TagMeta('dog', sly.TagValueType.NONE)
                meta_cat = sly.TagMeta('cat', sly.TagValueType.NONE)
                meta_turtle = sly.TagMeta('turtle', sly.TagValueType.NONE)

                tag_metas_animals = sly.TagMetaCollection([meta_dog, meta_cat, meta_turtle])

                metas_intersections = tag_metas.intersection(tag_metas_animals)
                print(metas_intersections.to_json())
                # Output: []

                # Let's add the potato TagMeta from another collection and compare them again
                tag_metas_animals = tag_metas_animals.add(meta_potato)

                metas_intersections = tag_metas.intersection(tag_metas_animals)
                print(metas_intersections.to_json())
                # Output: [
                #     {
                #         "name":"potato",
                #         "value_type":"none",
                #         "color":"#8A710F",
                #         "hotkey":"",
                #         "applicable_type":"all",
                #         "classes":[]
                #     }
                # ]

                # Difference, finds difference between collection and given list of TagMetas or TagMetaCollection
                meta_car = sly.TagMeta('car', sly.TagValueType.NONE)
                meta_bicycle = sly.TagMeta('bicycle', sly.TagValueType.NONE)

                tag_metas_vehicles = sly.TagMetaCollection([meta_car, meta_bicycle])

                meta_pedestrian = sly.TagMeta('pedestrian', sly.TagValueType.NONE)
                meta_road = sly.TagMeta('road', sly.TagValueType.NONE)

                difference = tag_metas_vehicles.difference([meta_pedestrian, meta_road])
                print(difference.to_json())
                # Output: [
                #     {
                #         "name":"car",
                #         "value_type":"none",
                #         "color":"#0F138A",
                #         "hotkey":"",
                #         "applicable_type":"all",
                #         "classes":[]
                #     },
                #     {
                #         "name":"bicycle",
                #         "value_type":"none",
                #         "color":"#0F8A25",
                #         "hotkey":"",
                #         "applicable_type":"all",
                #         "classes":[]
                #     }
                # ]

                # Merge, merges collection and given list of TagMetas
                tag_metas_vehicles = sly.TagMetaCollection([meta_car, meta_bicycle])
                tag_metas_merge = sly.TagMetaCollection([meta_pedestrian, meta_road])

                merged_collections = tag_metas_vehicles.merge(tag_metas_merge)
                print(merged_collections.to_json())
                # Output: [
                #     {
                #         "name":"pedestrian",
                #         "value_type":"none",
                #         "color":"#698A0F",
                #         "hotkey":"",
                #         "applicable_type":"all",
                #         "classes":[]
                #     },
                #     {
                #         "name":"road",
                #         "value_type":"none",
                #         "color":"#0F8A59",
                #         "hotkey":"",
                #         "applicable_type":"all",
                #         "classes":[]
                #     },
                #     {
                #         "name":"car",
                #         "value_type":"none",
                #         "color":"#0F138A",
                #         "hotkey":"",
                #         "applicable_type":"all",
                #         "classes":[]
                #     },
                #     {
                #         "name":"bicycle",
                #         "value_type":"none",
                #         "color":"#0F8A25",
                #         "hotkey":"",
                #         "applicable_type":"all",
                #         "classes":[]
                #     }
                # ]

                # Merge will raise ValueError if item name from given list is in collection but items in both are different
                meta_bicycle_1 = sly.TagMeta('bicycle', sly.TagValueType.NONE)
                meta_bicycle_2 = sly.TagMeta('bicycle', sly.TagValueType.ANY_STRING)

                tag_metas_1 = sly.TagMetaCollection([meta_bicycle_1])
                tag_metas_2 = sly.TagMetaCollection([meta_bicycle_2])

                test_merge = tag_metas_1.merge(tag_metas_2)
                # Output: ValueError: Error during merge for key 'bicycle': values are different

                # Let's try to create now a collection where TagMetas have identical names
                meta_cow = sly.TagMeta('cow', sly.TagValueType.NONE)
                meta_chicken = sly.TagMeta('cow', sly.TagValueType.NONE)

                tag_metas = sly.TagMetaCollection([meta_cow, meta_chicken])
                # Output: DuplicateKeyError: "Key 'cow' already exists"
        """
        super().__init__(items)

    def to_json(self) -> List[Dict]:
        """
        Convert the TagMetaCollection to a list of json dicts. Read more about `Supervisely format <https://docs.supervisely.com/data-organization/00_ann_format_navi>`_.

        :returns: List of dicts in json format
        :rtype: :class:`List[dict]`

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                tag_metas = tag_metas.to_json()
                # Output:
                # [
                #   {
                #    "name": "Weather",
                #    "value_type": "any_string",
                #    "color": "#8A620F",
                #    "hotkey": "",
                #    "applicable_type": "all",
                #    "classes": []
                #   },
                #   {
                #    "name": "Season",
                #    "value_type": "oneof_string",
                #    "color": "#700F8A",
                #    "values": ["Winter", 'Spring", "Summer", "Autumn"],
                #    "hotkey": "",
                #    "applicable_type": "all",
                #    "classes": []
                #   }
                # ]
        """
        return [tag_meta.to_json() for tag_meta in self]

    @classmethod
    def from_json(cls, data: List[Dict]) -> TagMetaCollection:
        """
        Convert a list with dicts in json format to TagMetaCollection. Read more about `Supervisely format <https://docs.supervisely.com/data-organization/00_ann_format_navi>`_.

        :param data: List with dicts in json format.
        :type data: List[dict]
        :returns: TagMetaCollection object
        :rtype: :class:`~supervisely.annotation.tag_meta_collection.TagMetaCollection`

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                data = [
                    {
                        "name":"Weather",
                        "value_type":"any_string",
                        "color":"#8A620F",
                        "hotkey":"",
                        "applicable_type":"all",
                        "classes":[]
                    },
                    {
                        "name":"Season",
                        "value_type": "oneof_string",
                                    "color": "#700F8A",
                                    "values": ["Winter", "Spring", "Summer", "Autumn"],
                        "hotkey":"",
                        "applicable_type":"all",
                        "classes":[]
                    }
                ]

                tag_metas = sly.TagMetaCollection.from_json(data)
        """
        tags = [TagMeta.from_json(tag_meta_json) for tag_meta_json in data]
        return cls(tags)

    def get_id_mapping(self, raise_if_no_id: Optional[bool] = False) -> Dict[int, TagMeta]:
        """
        Create dict matching TagMetas id to TagMeta.

        :param raise_if_no_id: Raise ValueError if where is TagMetas without id.
        :type raise_if_no_id: bool, optional
        :returns: Json format as a dict
        :rtype: dict
        :raises KeyError: if where is duplication of TagMetas id
        """
        res = {}
        without_id = []
        for tag_meta in self:
            if tag_meta.sly_id is not None:
                if tag_meta.sly_id in res:
                    raise KeyError(
                        f"TagMeta with id={tag_meta.sly_id} already exists (duplication). "
                        f"Please contact tech support"
                    )
                else:
                    res[tag_meta.sly_id] = tag_meta
            else:
                without_id.append(tag_meta)
        if len(without_id) > 0 and raise_if_no_id is True:
            raise ValueError("There are TagMetas without id")
        return res

    def __iter__(self) -> Iterator[TagMeta]:
        return next(self)

    def refresh_ids_from(self, tags: TagMetaCollection) -> None:
        """Update ids of TagMetas in collection from given collection.

        :param tags: TagMetaCollection to update ids from.
        :type tags: :class:`~supervisely.annotation.tag_meta_collection.TagMetaCollection`
        :returns: None
        :rtype: None
        """
        for new_tag in tags:
            my_tag = self.get(new_tag.name)
            if my_tag is None:
                continue
            my_tag._set_id(new_tag.sly_id)

    def get_by_id(self, tag_meta_id: int) -> TagMeta:
        """Return TagMeta with given id.

        :param tag_meta_id: TagMeta id to search for.
        :type tag_meta_id: int
        :returns: TagMeta with given id.
        :rtype: :class:`~supervisely.annotation.tag_meta.TagMeta`
        """
        for tag_meta in self:
            if tag_meta.sly_id == tag_meta_id:
                return tag_meta
        return None

    def get_tag_name_by_id(self, tag_meta_id: int) -> Optional[str]:
        for tag_meta in self:
            if tag_meta.sly_id == tag_meta_id:
                return tag_meta.name
        return None

    def merge(self, other: TagMetaCollection) -> TagMetaCollection:
        """
        Merge two TagMetaCollection objects.

        :param other: TagMetaCollection object.
        :type other:  :class:`~supervisely.annotation.tag_meta_collection.TagMetaCollection`
        :raises ValueError: if item name from given collection is in :class:`~supervisely.annotation.tag_meta_collection.TagMetaCollection` but TagMetas in both are different
        :returns: TagMetaCollection object
        :rtype: :class:`~supervisely.annotation.tag_meta_collection.TagMetaCollection`

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                item_cat = sly.TagMeta('cat', sly.TagValueType.NONE)
                item_turtle = sly.TagMeta('turtle', sly.TagValueType.ANY_STRING)
                collection = sly.TagMetaCollection([item_cat, item_turtle])

                item_dog = sly.TagMeta('dog', sly.TagValueType.NONE)
                item_turtle = sly.TagMeta('turtle', sly.TagValueType.ANY_STRING)
                other_collection = sly.TagMetaCollection([item_dog, item_turtle])

                merge = collection.merge(other_collection)
                print(merge.to_json())
                # Output: [
                #     {
                #         "name": "dog",
                #         "value_type": "none",
                #         "color": "#8A6C0F",
                #         "hotkey": "",
                #         "applicable_type": "all",
                #         "classes": []
                #     },
                #     {
                #         "name": "cat",
                #         "value_type": "none",
                #         "color": "#0F4A8A",
                #         "hotkey": "",
                #         "applicable_type": "all",
                #         "classes": []
                #     },
                #     {
                #         "name": "turtle",
                #         "value_type": "any_string",
                #         "color": "#4F0F8A",
                #         "hotkey": "",
                #         "applicable_type": "all",
                #         "classes": []
                #     }
                # ]
        """
        new_tags = []
        for other_tag in other.items():
            our_tag = self.get(other_tag.key())
            if our_tag is None:
                new_tags.append(other_tag)
            elif our_tag != other_tag:
                if (
                    our_tag.value_type == TagValueType.ONEOF_STRING
                    and other_tag.value_type == TagValueType.ONEOF_STRING
                ):
                    if Counter(our_tag.possible_values) == Counter(other_tag.possible_values):
                        continue
                raise ValueError(
                    "Error during merge for tag {!r}: values are different".format(other_tag.key())
                )
        return self.clone(new_tags + self.items())


def make_renamed_tag_metas(
    src_tag_metas: TagMetaCollection, renamer: Renamer, skip_missing: bool = False
) -> TagMetaCollection:
    """Returns a new TagMetaCollection with renamed TagMetas.

    :param src_tag_metas: TagMetaCollection to rename.
    :type src_tag_metas: :class:`~supervisely.annotation.tag_meta_collection.TagMetaCollection`
    :param renamer: Renamer object, which will handle renaming process.
    :type renamer: :class:`~supervisely.annotation.renamer.Renamer`
    :param skip_missing: If True, missing TagMetas will be skipped, otherwise KeyError will be raised.
    :type skip_missing: bool, optional
    :returns: New TagMetaCollection with renamed TagMetas.
    :rtype: :class:`~supervisely.annotation.tag_meta_collection.TagMetaCollection`
    """
    result_tags = []
    for src_tag in src_tag_metas:
        renamed_name = renamer.rename(src_tag.name)
        if renamed_name is not None:
            result_tags.append(src_tag.clone(name=renamed_name))
        elif not skip_missing:
            raise KeyError(
                "Tag meta named {} could not be mapped to a destination name.".format(src_tag.name)
            )
    return TagMetaCollection(items=result_tags)



from supervisely.geometry.geometry import Geometry
from supervisely.geometry.constants import LABELER_LOGIN, UPDATED_AT, CREATED_AT, ID, CLASS_ID
from supervisely.geometry.cuboid_3d import Vector3d
from typing import List, Union


class Polyline3D(Geometry):
    """3D polyline: sequence of connected line segments in 3D space. Immutable."""

    @staticmethod
    def geometry_name():
        """
        Returns the name of the geometry.

        :returns: name of the geometry
        :rtype: str
        """
        return "polyline_3d"

    def __init__(
        self,
        points: Union[List[float], List[Vector3d]],
        sly_id=None,
        class_id=None,
        labeler_login=None,
        updated_at=None,
        created_at=None,
    ):
        """
        Polyline3D geometry

        :param points: List of 3D point coordinates which define the polyline in 3D space.
        :type points: List[:class:`~supervisely.geometry.cuboid_3d.Vector3d`]
        :param sly_id: Polyline ID in Supervisely server.
        :type sly_id: int, optional
        :param class_id: ID of ObjClass to which Polyline belongs.
        :type class_id: int, optional
        :param labeler_login: Login of the user who created Polyline3D.
        :type labeler_login: str, optional
        :param updated_at: Date and Time when Polyline was modified last. Date Format: Year:Month:Day:Hour:Minute:Seconds. Example: '2021-01-22T19:37:50.158Z'.
        :type updated_at: str, optional
        :param created_at: Date and Time when Polyline was created. Date Format is the same as in "updated_at" parameter.
        :type created_at: str, optional

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                points = [[1, 2, 3], [4, 5, 6], [7, 8, 9]]
                figure = sly.Polyline(points)
        """
        if not isinstance(points[0], Vector3d):
            points = [Vector3d(point[0], point[1], point[2]) for point in points]
        super().__init__(
            sly_id=sly_id,
            class_id=class_id,
            labeler_login=labeler_login,
            updated_at=updated_at,
            created_at=created_at,
        )

        self._points = points

    @property
    def points(self):
        return self._points

    def to_json(self):
        points = [[point.x, point.y, point.z] for point in self._points]
        res = {"points": points}
        self._add_creation_info(res)
        return res

    @classmethod
    def from_json(cls, data):
        """
        Convert a json dict to Polyline3D.

        :param data: Polyline3D in json format as a dict.
        :type data: dict
        :returns: Polyline3D from json.
        :rtype: :class:`~supervisely.geometry.polyline_3d.Polyline3D`

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                figure_json = {
                    "points": {
                        [
                            [1, 2, 3],
                            [4, 5, 6],
                            [7, 8, 9]
                        ],
                    }
                }
                figure = sly.Polyline3D.from_json(figure_json)
        """
        if not data.get("points"):
            raise ValueError("Data dict must contain 'points' field!")
        points = data["points"]
        labeler_login = data.get(LABELER_LOGIN, None)
        updated_at = data.get(UPDATED_AT, None)
        created_at = data.get(CREATED_AT, None)
        sly_id = data.get(ID, None)
        class_id = data.get(CLASS_ID, None)
        return cls(
            points,
            sly_id=sly_id,
            class_id=class_id,
            labeler_login=labeler_login,
            updated_at=updated_at,
            created_at=created_at,
        )
